# Copyright (c) Xavier Figueroa
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List

from rdflib import Literal

from spdx import license, utils
from spdx.checksum import Checksum
from spdx.package import ExternalPackageRef
from spdx.relationship import Relationship
from spdx.utils import update_dict_item_with_new_item


class BaseWriter(object):
    """
    Base class for all Writer classes.
    Provide utility functions and stores shared fields.
    - document: spdx.document class. Source of data to be written
    - document_object: python dictionary representation of the entire spdx.document
    """

    def __init__(self, document):
        self.document = document
        self.document_object = dict()

    def license(self, license_field):
        """
        Return a string representation of a license or spdx.utils special object
        """
        if isinstance(
            license_field, (license.LicenseDisjunction, license.LicenseConjunction)
        ):
            return "({})".format(license_field)

        if isinstance(license_field, license.License):
            license_str = license_field.identifier.__str__()
        else:
            license_str = license_field.__str__()
        return license_str

    def checksum_to_dict(self, checksum_field: Checksum) -> Dict:
        """
        Return a dictionary representation of a checksum.Checksum object
        """
        return {'algorithm': checksum_field.identifier.name, 'checksumValue': checksum_field.value}

    def spdx_id(self, spdx_id_field):
        return spdx_id_field.__str__().split("#")[-1]


class CreationInfoWriter(BaseWriter):
    """
    Represent spdx.creationinfo as json-serializable objects
    """

    def __init__(self, document):
        super(CreationInfoWriter, self).__init__(document)

    def create_creation_info(self):
        creation_info_object = dict()
        creation_info = self.document.creation_info
        creation_info_object["creators"] = list(map(str, creation_info.creators))
        creation_info_object["created"] = creation_info.created_iso_format

        if creation_info.license_list_version:
            creation_info_object["licenseListVersion"] = "{0}.{1}".format(
                creation_info.license_list_version.major,
                creation_info.license_list_version.minor,
            )

        if creation_info.has_comment:
            creation_info_object["comment"] = creation_info.comment

        return creation_info_object


class PackageWriter(BaseWriter):
    """
    Represent spdx.package as python objects
    """

    def __init__(self, document):
        super(PackageWriter, self).__init__(document)

    def package_verification_code(self, package):
        """
        Represent the package verification code information as
        as python dictionary
        """

        package_verification_code_object = dict()

        package_verification_code_object["packageVerificationCodeValue"] = package.verif_code

        if package.verif_exc_files:
            package_verification_code_object[
                "packageVerificationCodeExcludedFiles"
            ] = package.verif_exc_files

        return package_verification_code_object

    @staticmethod
    def external_reference_as_dict(external_ref: ExternalPackageRef) -> dict:
        """
        Create a dictionary representation of the provided external reference, renaming the properties as they should
        appear in a json/yaml/xml document.
        """
        external_ref_dict = dict()
        external_ref_dict["referenceCategory"] = external_ref.category
        external_ref_dict["referenceType"] = external_ref.pkg_ext_ref_type
        external_ref_dict["referenceLocator"] = external_ref.locator
        if external_ref.comment:
            external_ref_dict["comment"] = external_ref.comment
        return external_ref_dict

    def create_package_info(self, package, annotations_by_spdx_id):
        package_object = dict()
        package_object["SPDXID"] = self.spdx_id(package.spdx_id)
        package_object["name"] = package.name
        package_object["downloadLocation"] = package.download_location.__str__()
        if package.files_analyzed is not None:
            package_object["filesAnalyzed"] = package.files_analyzed
        if package.files_analyzed is None or package.files_analyzed is True:
            if package.verif_code:
                package_object["packageVerificationCode"] = self.package_verification_code(
                    package
                )
            if package.has_optional_field("licenses_from_files"):
                package_object["licenseInfoFromFiles"] = list(
                    map(self.license, package.licenses_from_files)
                )
        if package.has_optional_field("conc_lics"):
            package_object["licenseConcluded"] = self.license(package.conc_lics)

        if package.has_optional_field("license_declared"):
            package_object["licenseDeclared"] = self.license(package.license_declared)

        if package.has_optional_field("cr_text"):
            package_object["copyrightText"] = package.cr_text.__str__()

        if package.has_optional_field("version"):
            package_object["versionInfo"] = package.version

        if package.has_optional_field("summary"):
            package_object["summary"] = package.summary

        if package.has_optional_field("attribution_text"):
            package_object["attributionTexts"] = [package.attribution_text]

        if package.has_optional_field("source_info"):
            package_object["sourceInfo"] = package.source_info

        if package.has_optional_field("file_name"):
            package_object["packageFileName"] = package.file_name

        if package.has_optional_field("supplier"):
            package_object["supplier"] = package.supplier.to_value()

        if package.has_optional_field("originator"):
            package_object["originator"] = package.originator.to_value()

        for checksum in package.checksums.values():
            package_object.setdefault("checksums", []).append(self.checksum_to_dict(checksum))

        if package.has_optional_field("description"):
            package_object["description"] = package.description

        if package.has_optional_field("comment"):
            package_object["comment"] = package.comment

        if package.has_optional_field("license_comment"):
            package_object["licenseComments"] = package.license_comment

        if package.has_optional_field("homepage"):
            package_object["homepage"] = package.homepage.__str__()

        if package.has_optional_field("primary_package_purpose"):
            package_object["primaryPackagePurpose"] = package.primary_package_purpose.name.replace("_", "-")

        if package.has_optional_field("release_date"):
            package_object["releaseDate"] = utils.datetime_iso_format(package.release_date)

        if package.has_optional_field("built_date"):
            package_object["builtDate"] = utils.datetime_iso_format(package.built_date)

        if package.has_optional_field("valid_until_date"):
            package_object["validUntilDate"] = utils.datetime_iso_format(package.valid_until_date)

        if package.has_optional_field("pkg_ext_refs"):
            package_object["externalRefs"] = [self.external_reference_as_dict(external_ref) for external_ref in
                                              package.pkg_ext_refs]
        if package.spdx_id in annotations_by_spdx_id:
            package_object["annotations"] = annotations_by_spdx_id[package.spdx_id]

        return package_object


class FileWriter(BaseWriter):
    """
    Represent spdx.file as json-serializable objects
    """

    def __init__(self, document):
        super(FileWriter, self).__init__(document)

    def create_artifact_info(self, file):
        """
        Create the artifact json-serializable representation from a spdx.file.File object
        """
        artifact_of_objects = []

        for i in range(len(file.artifact_of_project_name)):
            artifact_of_object = dict()
            artifact_of_object["name"] = file.artifact_of_project_name[i].__str__()
            artifact_of_object["homePage"] = file.artifact_of_project_home[i].__str__()
            artifact_of_object["projectUri"] = file.artifact_of_project_uri[i].__str__()
            artifact_of_objects.append(artifact_of_object)

        return artifact_of_objects

    def create_file_info(self, file, annotations_by_spdx_id):
        file_object = dict()

        file_object["fileName"] = file.name
        file_object["SPDXID"] = self.spdx_id(file.spdx_id)
        for checksum in file.checksums.values():
            file_object.setdefault("checksums", []).append(self.checksum_to_dict(checksum))
        if file.has_optional_field("conc_lics"):
            file_object["licenseConcluded"] = self.license(file.conc_lics)

        if file.has_optional_field("licenses_in_file"):
            file_object["licenseInfoInFiles"] = list(
                map(self.license, file.licenses_in_file)
            )

        if file.has_optional_field("copyright"):
            file_object["copyrightText"] = file.copyright.__str__()

        if file.has_optional_field("comment"):
            file_object["comment"] = file.comment

        if file.has_optional_field("file_types"):
            types = []
            for file_type in file.file_types:
                types.append(file_type.name)
            file_object["fileTypes"] = types

        if file.has_optional_field("license_comment"):
            file_object["licenseComments"] = file.license_comment

        if file.has_optional_field("attribution_text"):
            file_object["attributionTexts"] = [file.attribution_text]

        if file.has_optional_field("notice"):
            file_object["noticeText"] = file.notice

        if file.contributors:
            file_object["fileContributors"] = file.contributors

        if file.dependencies:
            file_object["fileDependencies"] = file.dependencies

        valid_artifacts = (
            file.artifact_of_project_name
            and len(file.artifact_of_project_name)
            == len(file.artifact_of_project_home)
            and len(file.artifact_of_project_home)
            == len(file.artifact_of_project_uri)
        )

        if valid_artifacts:
            file_object["artifactOf"] = self.create_artifact_info(file)

        if file.spdx_id in annotations_by_spdx_id:
            file_object["annotations"] = annotations_by_spdx_id[file.spdx_id]

        return file_object


class ReviewInfoWriter(BaseWriter):
    """
    Represent spdx.review as json-serializable objects
    """

    def __init__(self, document):
        super(ReviewInfoWriter, self).__init__(document)

    def create_review_info(self):
        review_info_objects = []
        reviews = self.document.reviews

        for review in reviews:
            review_object = dict()
            review_object["reviewer"] = review.reviewer.__str__()
            review_object["reviewDate"] = review.review_date_iso_format
            if review.has_comment:
                review_object["comment"] = review.comment

            review_info_objects.append(review_object)

        return review_info_objects


class AnnotationInfoWriter(BaseWriter):
    """
    Represent spdx.annotation as json-serializable objects
    """

    def __init__(self, document):
        super(AnnotationInfoWriter, self).__init__(document)

    def create_annotations_by_spdx_id(self) -> Dict:
        """
        Create a dict with annotations_by_spdx_id and use the spdx_id of the element that is annotated as key.
        These keys are then used to attach the annotation to the corresponding SPDX element.
        """
        annotations_by_spdx_id = dict()

        if not self.document.annotations:
            return annotations_by_spdx_id

        for annotation in self.document.annotations:
            annotation_object = dict()
            annotation_object["annotator"] = annotation.annotator.__str__()
            annotation_object["annotationDate"] = annotation.annotation_date_iso_format
            annotation_object["annotationType"] = annotation.annotation_type
            annotation_object["comment"] = annotation.comment

            annotation_spdx_id = self.spdx_id(annotation.spdx_id)
            if annotation_spdx_id not in annotations_by_spdx_id:
                annotations_by_spdx_id[annotation_spdx_id] = [annotation_object]
            else:
                annotations_by_spdx_id[annotation_spdx_id].append(annotation_object)

        return annotations_by_spdx_id


class RelationshipInfoWriter(BaseWriter):
    """
    Represent spdx.relationship as json-serializable objects
    """

    def __init__(self, document):
        super(RelationshipInfoWriter, self).__init__(document)

    def create_relationship_info(self, relationship: Relationship):
        relationship_object = dict()
        relationship_object["spdxElementId"] = relationship.spdx_element_id
        relationship_object[
            "relatedSpdxElement"
        ] = relationship.related_spdx_element
        relationship_object["relationshipType"] = relationship.relationship_type
        if relationship.has_comment:
            relationship_object["comment"] = relationship.relationship_comment

        return relationship_object


class SnippetWriter(BaseWriter):
    """
    Represent spdx.snippet as json-serializable objects
    """

    def __init__(self, document):
        super(SnippetWriter, self).__init__(document)

    def create_snippet_info(self, annotations_by_spdx_id):
        snippet_info_objects = []
        snippets = self.document.snippet

        for snippet in snippets:
            snippet_from_file_spdx_id = self.spdx_id(snippet.snip_from_file_spdxid)
            snippet_object = dict()
            snippet_object["SPDXID"] = self.spdx_id(snippet.spdx_id)
            snippet_object["snippetFromFile"] = snippet_from_file_spdx_id

            if snippet.has_optional_field("copyright"):
                snippet_object["copyrightText"] = snippet.copyright

            if snippet.has_optional_field("conc_lics"):
                snippet_object["licenseConcluded"] = self.license(snippet.conc_lics)

            if snippet.has_optional_field("licenses_in_snippet"):
                snippet_object["licenseInfoInSnippets"] = list(
                    map(self.license, snippet.licenses_in_snippet)
                )
            byte_range = {"endPointer": {"offset": snippet.byte_range[1], "reference": snippet_from_file_spdx_id},
                          "startPointer": {"offset": snippet.byte_range[0], "reference": snippet_from_file_spdx_id}}
            snippet_object["ranges"] = [byte_range]

            if snippet.has_optional_field("name"):
                snippet_object["name"] = snippet.name

            if snippet.has_optional_field("comment"):
                snippet_object["comment"] = snippet.comment

            if snippet.has_optional_field("attribution_text"):
                snippet_object["attributionTexts"] = [snippet.attribution_text]

            if snippet.has_optional_field("license_comment"):
                snippet_object["licenseComments"] = snippet.license_comment

            if snippet.spdx_id in annotations_by_spdx_id:
                snippet_object["annotations"] = annotations_by_spdx_id[snippet.spdx_id]

            if snippet.has_optional_field("line_range"):
                line_range = {
                    "endPointer": {"lineNumber": snippet.line_range[1], "reference": snippet_from_file_spdx_id},
                    "startPointer": {"lineNumber": snippet.line_range[0], "reference": snippet_from_file_spdx_id}}
                snippet_object["ranges"].append(line_range)

            snippet_info_objects.append(snippet_object)

        return snippet_info_objects


class ExtractedLicenseWriter(BaseWriter):
    """
    Represent spdx.document.ExtractedLicense as json-serializable objects
    """

    def __init__(self, document):
        super(ExtractedLicenseWriter, self).__init__(document)

    def create_extracted_license(self):
        extracted_license_objects = []
        unique_extracted_licenses = {}
        for lic in self.document.extracted_licenses:
            if lic.identifier not in unique_extracted_licenses.keys():
                unique_extracted_licenses[lic.identifier] = lic

        for extracted_license in unique_extracted_licenses.values():
            extracted_license_object = dict()

            if isinstance(extracted_license.identifier, Literal):
                extracted_license_object[
                    "licenseId"
                ] = extracted_license.identifier.toPython()
            else:
                extracted_license_object["licenseId"] = extracted_license.identifier

            if isinstance(extracted_license.text, Literal):
                extracted_license_object[
                    "extractedText"
                ] = extracted_license.text.toPython()
            else:
                extracted_license_object["extractedText"] = extracted_license.text

            if extracted_license.full_name:
                if isinstance(extracted_license.full_name, Literal):
                    extracted_license_object[
                        "name"
                    ] = extracted_license.full_name.toPython()
                else:
                    extracted_license_object["name"] = extracted_license.full_name

            if extracted_license.cross_ref:
                if isinstance(extracted_license.cross_ref, Literal):
                    extracted_license_object[
                        "seeAlsos"
                    ] = extracted_license.cross_ref.toPython()
                else:
                    extracted_license_object["seeAlsos"] = extracted_license.cross_ref

            if extracted_license.comment:
                if isinstance(extracted_license.comment, Literal):
                    extracted_license_object[
                        "comment"
                    ] = extracted_license.comment.toPython()
                else:
                    extracted_license_object["comment"] = extracted_license.comment

            extracted_license_objects.append(extracted_license_object)

        return extracted_license_objects


class Writer(
    CreationInfoWriter,
    ReviewInfoWriter,
    FileWriter,
    PackageWriter,
    AnnotationInfoWriter,
    RelationshipInfoWriter,
    SnippetWriter,
    ExtractedLicenseWriter,
):
    """
    Wrapper for the other writers.
    Represent a whole SPDX Document as json-serializable objects to then
    be written as json or yaml files.
    """

    def __init__(self, document):
        self.doc_spdx_id = self.spdx_id(document.spdx_id)
        super(Writer, self).__init__(document)

    def create_ext_document_references(self):
        """
        Create the External Document References json-serializable representation
        """
        ext_document_references_field = self.document.ext_document_references
        ext_document_reference_objects = []
        for ext_document_reference in ext_document_references_field:
            ext_document_reference_object = dict()
            ext_document_reference_object[
                "externalDocumentId"
            ] = ext_document_reference.external_document_id
            ext_document_reference_object[
                "spdxDocument"
            ] = ext_document_reference.spdx_document_uri

            ext_document_reference_object["checksum"] = self.checksum_to_dict(
                ext_document_reference.checksum
            )

            ext_document_reference_objects.append(ext_document_reference_object)

        return ext_document_reference_objects

    def create_relationships(self) -> List[Dict]:
        packages_spdx_ids = [package.spdx_id for package in self.document.packages]
        files_spdx_ids = [file.spdx_id for file in self.document.files]
        # we take the package_objects from document_object if any exist because we will modify them to add
        # jsonyamlxml-specific fields
        if "packages" in self.document_object:
            packages_by_spdx_id = {package["SPDXID"]: package for package in self.document_object["packages"]}
        else:
            packages_by_spdx_id = {}

        relationship_objects = []
        for relationship in self.document.relationships:
            if relationship.relationship_type == "CONTAINS" and relationship.spdx_element_id in packages_spdx_ids \
                    and relationship.related_spdx_element in files_spdx_ids:
                update_dict_item_with_new_item(packages_by_spdx_id[relationship.spdx_element_id], "hasFiles",
                                               relationship.related_spdx_element)
                if relationship.has_comment:
                    relationship_objects.append(self.create_relationship_info(relationship))

            elif relationship.relationship_type == "CONTAINED_BY" and relationship.spdx_element_id in files_spdx_ids \
                    and relationship.related_spdx_element in packages_spdx_ids:
                update_dict_item_with_new_item(packages_by_spdx_id[relationship.related_spdx_element],
                                                    "hasFiles", relationship.spdx_element_id)
                if relationship.has_comment:
                    relationship_objects.append(self.create_relationship_info(relationship))

            elif relationship.relationship_type == "DESCRIBES" and relationship.spdx_element_id == self.document.spdx_id:
                update_dict_item_with_new_item(self.document_object, "documentDescribes",
                                               relationship.related_spdx_element)
                if relationship.has_comment:
                    relationship_objects.append(self.create_relationship_info(relationship))

            elif relationship.relationship_type == "DESCRIBED_BY" and relationship.related_spdx_element == self.document.spdx_id:
                update_dict_item_with_new_item(self.document_object, "documentDescribes",
                                               relationship.spdx_element_id)
                if relationship.has_comment:
                    relationship_objects.append(self.create_relationship_info(relationship))

            else:
                relationship_objects.append(self.create_relationship_info(relationship))

        return relationship_objects

    def create_document(self):
        self.document_object = dict()

        self.document_object["spdxVersion"] = self.document.version.__str__()
        self.document_object["documentNamespace"] = self.document.namespace.__str__()
        self.document_object["creationInfo"] = self.create_creation_info()
        self.document_object["dataLicense"] = self.license(self.document.data_license)
        self.document_object["SPDXID"] = self.doc_spdx_id
        self.document_object["name"] = self.document.name
        annotations_by_spdx_id = self.create_annotations_by_spdx_id()

        unique_doc_packages = {}
        for doc_package in self.document.packages:
            if doc_package.spdx_id not in unique_doc_packages.keys():
                unique_doc_packages[doc_package.spdx_id] = doc_package
        if unique_doc_packages:
            package_objects = []
            for package in unique_doc_packages.values():
                package_info_object = self.create_package_info(package, annotations_by_spdx_id)
                package_objects.append(package_info_object)
            self.document_object["packages"] = package_objects
        if self.document.files:
            file_objects = []
            for file in self.document.files:
                file_object = self.create_file_info(file, annotations_by_spdx_id)
                file_objects.append(file_object)
            self.document_object["files"] = file_objects

        if self.document.has_comment:
            self.document_object["comment"] = self.document.comment

        if self.document.ext_document_references:
            self.document_object[
                "externalDocumentRefs"
            ] = self.create_ext_document_references()

        if self.document.extracted_licenses:
            self.document_object[
                "hasExtractedLicensingInfos"
            ] = self.create_extracted_license()

        if self.document.reviews:
            self.document_object["reviewers"] = self.create_review_info()

        if self.document.snippet:
            self.document_object["snippets"] = self.create_snippet_info(annotations_by_spdx_id)

        if self.doc_spdx_id in annotations_by_spdx_id:
            self.document_object["annotations"] = annotations_by_spdx_id[self.doc_spdx_id]

        if self.document.relationships:
            relationship_objects = self.create_relationships()
            if relationship_objects:
                self.document_object["relationships"] = relationship_objects

        return self.document_object



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Dict, List

from spdx import annotation
from spdx import creationinfo
from spdx import file
from spdx import license
from spdx import package
from spdx import review
from spdx import snippet
from spdx import utils
from spdx import version
from spdx.checksum import Checksum
from spdx.document import ExternalDocumentRef, Document
from spdx.package import PackagePurpose
from spdx.parsers import validations
from spdx.parsers.builderexceptions import CardinalityError
from spdx.parsers.builderexceptions import OrderError
from spdx.parsers.builderexceptions import SPDXValueError
from spdx.relationship import Relationship


def str_from_text(text) -> str:
    """
    Return content of a free form text block as a string.
    """
    REGEX = re.compile("<text>((.|\n)+)</text>", re.UNICODE)
    match = REGEX.match(text)
    if match:
        return match.group(1)
    elif isinstance(text, str):
        return text
    else:
        return None


class DocBuilder(object):
    """
    Set the fields of the top level document model.
    """

    VERS_STR_REGEX = re.compile(r"SPDX-(\d+)\.(\d+)", re.UNICODE)

    def __init__(self):
        # FIXME: this state does not make sense
        self.reset_document()

    def set_doc_version(self, doc, value):
        """
        Set the document version.
        Raise SPDXValueError if malformed value.
        Raise CardinalityError if already defined.
        """
        if self.doc_version_set:
            raise CardinalityError("Document::Version")

        m = self.VERS_STR_REGEX.match(value)
        if m is None:
            raise SPDXValueError("Document::Version")

        self.doc_version_set = True
        doc.version = version.Version(
            major=int(m.group(1)), minor=int(m.group(2))
        )
        return True

    def set_doc_data_lics(self, doc, lics):
        """
        Set the document data license.
        Raise value error if malformed value
        Raise CardinalityError if already defined.
        """
        if self.doc_data_lics_set:
            raise CardinalityError("Document::DataLicense")

        if not validations.validate_data_lics(lics):
            raise SPDXValueError("Document::DataLicense")

        self.doc_data_lics_set = True
        doc.data_license = license.License.from_identifier(lics)
        return True

    def set_doc_name(self, doc, name):
        """
        Set the document name.
        Raise CardinalityError if already defined.
        """
        if self.doc_name_set:
            raise CardinalityError("Document::Name")

        self.doc_name_set = True
        doc.name = name
        return True

    def set_doc_spdx_id(self, doc, doc_spdx_id_line):
        """
        Set the document SPDX Identifier.
        Raise value error if malformed value.
        Raise CardinalityError if already defined.
        """
        if self.doc_spdx_id_set:
            raise CardinalityError("Document::SPDXID")

        if not doc_spdx_id_line == "SPDXRef-DOCUMENT":
            raise SPDXValueError("Document::SPDXID")

        doc.spdx_id = doc_spdx_id_line
        self.doc_spdx_id_set = True
        return True

    def set_doc_comment(self, doc, comment):
        """
        Set document comment.
        Raise CardinalityError if comment already set.
        Raise SPDXValueError if comment is not free form text or single line of text.
        """
        if self.doc_comment_set:
            raise CardinalityError("Document::Comment")

        if not validations.validate_doc_comment(comment):
            raise SPDXValueError("Document::Comment")

        self.doc_comment_set = True
        doc.comment = str_from_text(comment)
        return True

    def set_doc_namespace(self, doc, namespace):
        """
        Set the document namespace.
        Raise SPDXValueError if malformed value.
        Raise CardinalityError if already defined.
        """
        if self.doc_namespace_set:
            raise CardinalityError("Document::Namespace")

        if not validations.validate_doc_namespace(namespace):
            raise SPDXValueError("Document::Namespace")

        self.doc_namespace_set = True
        doc.namespace = namespace
        return True

    def reset_document(self):
        """
        Reset the state to allow building new documents
        """
        # FIXME: this state does not make sense
        self.doc_version_set = False
        self.doc_comment_set = False
        self.doc_namespace_set = False
        self.doc_data_lics_set = False
        self.doc_name_set = False
        self.doc_spdx_id_set = False


class ExternalDocumentRefBuilder(object):
    def set_ext_doc_id(self, doc, ext_doc_id):
        """
        Set the `external_document_id` attribute of the `ExternalDocumentRef` object.
        """
        doc.add_ext_document_reference(
            ExternalDocumentRef(external_document_id=ext_doc_id)
        )

    def set_spdx_doc_uri(self, doc, spdx_doc_uri):
        """
        Set the `spdx_document_uri` attribute of the `ExternalDocumentRef` object.
        """
        if not validations.validate_doc_namespace(spdx_doc_uri):
            raise SPDXValueError("Document::ExternalDocumentRef")

        doc.ext_document_references[-1].spdx_document_uri = spdx_doc_uri

    def set_chksum(self, doc, chksum):
        """
        Set the `check_sum` attribute of the `ExternalDocumentRef` object.
        """
        doc.ext_document_references[-1].checksum = Checksum.checksum_from_string(chksum)

    def add_ext_doc_refs(self, doc, ext_doc_id, spdx_doc_uri, chksum):
        self.set_ext_doc_id(doc, ext_doc_id)
        self.set_spdx_doc_uri(doc, spdx_doc_uri)
        self.set_chksum(doc, chksum)


class EntityBuilder(object):
    tool_re = re.compile(r"Tool:\s*(.+)", re.UNICODE)
    person_re = re.compile(r"Person:\s*(([^(])+)(\((.*)\))?", re.UNICODE)
    org_re = re.compile(r"Organization:\s*(([^(])+)(\((.*)\))?", re.UNICODE)
    PERSON_NAME_GROUP = 1
    PERSON_EMAIL_GROUP = 4
    ORG_NAME_GROUP = 1
    ORG_EMAIL_GROUP = 4
    TOOL_NAME_GROUP = 1

    def build_tool(self, doc, entity):
        """
        Build a tool object out of a string representation.
        Return built tool.
        Raise SPDXValueError if failed to extract tool name or name is malformed
        """
        match = self.tool_re.match(entity)
        if not match or not validations.validate_tool_name(match.group(self.TOOL_NAME_GROUP)):
            raise SPDXValueError("Failed to extract tool name")

        name = match.group(self.TOOL_NAME_GROUP)
        return creationinfo.Tool(name)

    def build_org(self, doc, entity):
        """
        Build an organization object of of a string representation.
        Return built organization.
        Raise SPDXValueError if failed to extract name.
        """
        match = self.org_re.match(entity)
        if not match or not validations.validate_org_name(match.group(self.ORG_NAME_GROUP)):
            raise SPDXValueError("Failed to extract Organization name")

        name = match.group(self.ORG_NAME_GROUP).strip()
        email = match.group(self.ORG_EMAIL_GROUP)
        if (email is not None) and (len(email) != 0):
            return creationinfo.Organization(name=name, email=email.strip())
        else:
            return creationinfo.Organization(name=name, email=None)

    def build_person(self, doc, entity):
        """
        Build an organization object of of a string representation.
        Return built organization. Raise SPDXValueError if failed to extract name.
        """
        match = self.person_re.match(entity)
        if not match or not validations.validate_person_name(match.group(self.PERSON_NAME_GROUP)):
            raise SPDXValueError("Failed to extract person name")

        name = match.group(self.PERSON_NAME_GROUP).strip()
        email = match.group(self.PERSON_EMAIL_GROUP)
        if (email is not None) and (len(email) != 0):
            return creationinfo.Person(name=name, email=email.strip())
        else:
            return creationinfo.Person(name=name, email=None)


class CreationInfoBuilder(object):
    def __init__(self):
        # FIXME: this state does not make sense
        self.reset_creation_info()

    def add_creator(self, doc, creator):
        """
        Add a creator to the document's creation info.
        Return true if creator is valid.
        Creator must be built by an EntityBuilder.
        Raise SPDXValueError if not a creator type.
        """
        if not validations.validate_creator(creator):
            raise SPDXValueError("CreationInfo::Creator")

        doc.creation_info.add_creator(creator)
        return True

    def set_created_date(self, doc, created):
        """
        Set created date.
        Raise CardinalityError if created date already set.
        Raise SPDXValueError if created is not a date.
        """
        if self.created_date_set:
            raise CardinalityError("CreationInfo::Created")

        date = utils.datetime_from_iso_format(created)
        if date is None:
            raise SPDXValueError("CreationInfo::Date")

        self.created_date_set = True
        doc.creation_info.created = date
        return True

    def set_creation_comment(self, doc, comment):
        """
        Set creation comment.
        Raise CardinalityError if comment already set.
        Raise SPDXValueError if not free form text or single line of text.
        """
        if self.creation_comment_set:
            raise CardinalityError("CreationInfo::Comment")

        if not validations.validate_creation_comment(comment):
            raise SPDXValueError("CreationInfo::Comment")

        self.creation_comment_set = True
        doc.creation_info.comment = str_from_text(comment)
        return True

    def set_lics_list_ver(self, doc, value):
        """
        Set the license list version.
        Raise CardinalityError if already set.
        Raise SPDXValueError if incorrect value.
        """
        if self.lics_list_ver_set:
            raise CardinalityError("CreationInfo::LicenseListVersion")

        vers = version.Version.from_str(value)
        if vers is None:
            raise SPDXValueError("CreationInfo::LicenseListVersion")

        self.lics_list_ver_set = True
        doc.creation_info.license_list_version = vers
        return True

    def reset_creation_info(self):
        """
        Reset builder state to allow building new creation info.
        """
        # FIXME: this state does not make sense
        self.created_date_set = False
        self.creation_comment_set = False
        self.lics_list_ver_set = False


class ReviewBuilder(object):
    def __init__(self):
        # FIXME: this state does not make sense
        self.reset_reviews()

    def reset_reviews(self):
        """
        Reset the builder's state to allow building new reviews.
        """
        # FIXME: this state does not make sense
        self.review_date_set = False
        self.review_comment_set = False

    def add_reviewer(self, doc, reviewer):
        """
        Adds a reviewer to the SPDX Document.
        Reviewer is an entity created by an EntityBuilder.
        Raise SPDXValueError if not a valid reviewer type.
        """
        # Each reviewer marks the start of a new review object.
        # FIXME: this state does not make sense
        self.reset_reviews()
        if not validations.validate_reviewer(reviewer):
            raise SPDXValueError("Review::Reviewer")

        doc.add_review(review.Review(reviewer=reviewer))
        return True

    def add_review_date(self, doc, reviewed):
        """
        Set the review date.
        Raise CardinalityError if already set.
        Raise OrderError if no reviewer defined before.
        Raise SPDXValueError if invalid reviewed value.
        """
        if len(doc.reviews) == 0:
            raise OrderError("Review::ReviewDate")

        if self.review_date_set:
            raise CardinalityError("Review::ReviewDate")

        date = utils.datetime_from_iso_format(reviewed)
        if date is None:
            raise SPDXValueError("Review::ReviewDate")

        self.review_date_set = True
        doc.reviews[-1].review_date = date
        return True

    def add_review_comment(self, doc, comment):
        """
        Set the review comment.
        Raise CardinalityError if already set.
        Raise OrderError if no reviewer defined before.
        Raise SPDXValueError if comment is not free form text or single line of text.
        """
        if len(doc.reviews) == 0:
            raise OrderError("ReviewComment")

        if self.review_comment_set:
            raise CardinalityError("ReviewComment")

        if not validations.validate_review_comment(comment):
            raise SPDXValueError("ReviewComment::Comment")

        self.review_comment_set = True
        doc.reviews[-1].comment = str_from_text(comment)
        return True


class AnnotationBuilder(object):
    def __init__(self):
        # FIXME: this state does not make sense
        self.reset_annotations()

    def reset_annotations(self):
        """
        Reset the builder's state to allow building new annotations.
        """
        # FIXME: this state does not make sense
        self.annotation_date_set = False
        self.annotation_comment_set = False
        self.annotation_type_set = False
        self.annotation_spdx_id_set = False

    def add_annotator(self, doc, annotator):
        """
        Add an annotator to the SPDX Document.
        Annotator is an entity created by an EntityBuilder.
        Raise SPDXValueError if not a valid annotator type.
        """
        # Each annotator marks the start of a new annotation object.
        # FIXME: this state does not make sense
        self.reset_annotations()
        if not validations.validate_annotator(annotator):
            raise SPDXValueError("Annotation::Annotator")

        doc.add_annotation(annotation.Annotation(annotator=annotator))
        return True

    def add_annotation_date(self, doc, annotation_date):
        """
        Set the annotation date.
        Raise CardinalityError if already set.
        Raise OrderError if no annotator defined before.
        Raise SPDXValueError if invalid value.
        """
        if len(doc.annotations) == 0:
            raise OrderError("Annotation::AnnotationDate")

        if self.annotation_date_set:
            raise CardinalityError("Annotation::AnnotationDate")

        date = utils.datetime_from_iso_format(annotation_date)
        if date is None:
            raise SPDXValueError("Annotation::AnnotationDate")

        self.annotation_date_set = True
        doc.annotations[-1].annotation_date = date
        return True

    def add_annotation_comment(self, doc, comment):
        """
        Set the annotation comment.
        Raise CardinalityError if already set.
        Raise OrderError if no annotator defined before.
        Raise SPDXValueError if comment is not free form text or single line of text.
        """
        if len(doc.annotations) == 0:
            raise OrderError("AnnotationComment::Comment")

        if self.annotation_comment_set:
            raise CardinalityError("AnnotationComment::Comment")

        if not validations.validate_annotation_comment(comment):
            raise SPDXValueError("AnnotationComment::Comment")

        self.annotation_comment_set = True
        doc.annotations[-1].comment = str_from_text(comment)
        return True

    def add_annotation_type(self, doc, annotation_type):
        """
        Set the annotation type.
        Raise CardinalityError if already set.
        Raise OrderError if no annotator defined before.
        Raise SPDXValueError if invalid value.
        """
        if len(doc.annotations) == 0:
            raise OrderError("Annotation::AnnotationType")

        if self.annotation_type_set:
            raise CardinalityError("Annotation::AnnotationType")

        if not validations.validate_annotation_type(annotation_type):
            raise SPDXValueError("Annotation::AnnotationType")

        self.annotation_type_set = True
        doc.annotations[-1].annotation_type = annotation_type
        return True

    def set_annotation_spdx_id(self, doc, spdx_id):
        """
        Set the annotation SPDX Identifier.
        Raise CardinalityError if already set.
        Raise OrderError if no annotator defined before.
        """
        if len(doc.annotations) == 0:
            raise OrderError("Annotation::SPDXREF")

        if self.annotation_spdx_id_set:
            raise CardinalityError("Annotation::SPDXREF")

        self.annotation_spdx_id_set = True
        doc.annotations[-1].spdx_id = spdx_id
        return True


class RelationshipBuilder(object):
    def __init__(self):
        # FIXME: this state does not make sense
        self.reset_relationship()

    def reset_relationship(self):
        """
        Reset the builder's state to allow building new relationships.
        """
        # FIXME: this state does not make sense
        self.relationship_comment_set = False

    def add_relationship(self, doc: Document, relationship_term: str) -> bool:
        """
        Raise SPDXValueError if type is unknown.
        """
        self.reset_relationship()
        relationship_to_add = Relationship(relationship_term)
        existing_relationships: List[Relationship] = doc.relationships

        if relationship_to_add not in existing_relationships:
            doc.add_relationship(relationship_to_add)
            return True

        existing_relationship: Relationship = existing_relationships[existing_relationships.index(relationship_to_add)]

        # If the relationship already exists without comment, we remove the old one and re-append it at the end. This
        # allows to add a comment to the relationship (since a comment will always be added to the latest
        # relationship). If an equal relationship with comment already exists, we ignore the new relationship.
        if not existing_relationship.has_comment:
            existing_relationships.remove(relationship_to_add)
            doc.add_relationship(relationship_to_add)
            return True

        return False

    def add_relationship_comment(self, doc: Document, comment: str) -> bool:
        """
        Set the annotation comment.
        Raise CardinalityError if already set.
        Raise OrderError if no relationship defined before it.
        Raise SPDXValueError if comment is not free form text or single line of text.
        """
        if len(doc.relationships) == 0:
            raise OrderError("RelationshipComment::Comment")

        if self.relationship_comment_set:
            raise CardinalityError("RelationshipComment::Comment")

        if not validations.validate_relationship_comment(comment):
            raise SPDXValueError("RelationshipComment::Comment")

        self.relationship_comment_set = True
        doc.relationships[-1].comment = str_from_text(comment)
        return True


class PackageBuilder(object):
    VERIF_CODE_REGEX = re.compile(r"([0-9a-f]+)\s*(\(\s*(.+)\))?", re.UNICODE)
    VERIF_CODE_CODE_GRP = 1
    VERIF_CODE_EXC_FILES_GRP = 3

    def __init__(self):
        # FIXME: this state does not make sense
        self.reset_package()

    def reset_package(self):
        """Resets the builder's state in order to build new packages."""
        # FIXME: this state does not make sense
        self.package_set = False
        self.package_spdx_id_set = False
        self.package_vers_set = False
        self.package_file_name_set = False
        self.package_supplier_set = False
        self.package_originator_set = False
        self.package_down_location_set = False
        self.package_files_analyzed_set = False
        self.package_home_set = False
        self.package_verif_set = False
        self.package_chk_sum_set = False
        self.package_source_info_set = False
        self.package_conc_lics_set = False
        self.package_license_declared_set = False
        self.package_license_comment_set = False
        self.package_cr_text_set = False
        self.package_summary_set = False
        self.package_desc_set = False
        self.package_comment_set = False
        self.package_primary_purpose_set = False
        self.package_built_date_set = False
        self.package_release_date_set = False
        self.package_valid_until_date_set = False
        # self.package_attribution_text_set = False
        self.pkg_ext_comment_set = False

    def create_package(self, doc, name):
        """
        Create a package for the SPDX Document.
        name - any string.
        Raise CardinalityError if package already defined.
        """
        self.reset_package()
        self.package_set = True
        doc.add_package(package.Package(name=name))
        return True

    def set_pkg_spdx_id(self, doc, spdx_id):
        """
        Set the Package SPDX Identifier.
        Raise SPDXValueError if malformed value.
        Raise CardinalityError if already defined.
        """
        self.assert_package_exists()
        if self.package_spdx_id_set:
            raise CardinalityError("Package::SPDXID")

        if not validations.validate_pkg_spdx_id(spdx_id):
            raise SPDXValueError("Package::SPDXID")

        self.package_spdx_id_set = True
        doc.packages[-1].spdx_id = spdx_id
        return True

    def set_pkg_vers(self, doc, version):
        """
        Set package version, if not already set.
        version - Any string.
        Raise CardinalityError if already has a version.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if self.package_vers_set:
            raise CardinalityError("Package::Version")

        self.package_vers_set = True
        doc.packages[-1].version = version
        return True

    def set_pkg_file_name(self, doc, name):
        """
        Set the package file name, if not already set.
        name - Any string.
        Raise CardinalityError if already has a file_name.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if self.package_file_name_set:
            raise CardinalityError("Package::FileName")

        self.package_file_name_set = True
        doc.packages[-1].file_name = name
        return True

    def set_pkg_supplier(self, doc, entity):
        """
        Set the package supplier, if not already set.
        entity - Organization, Person or NoAssert.
        Raise CardinalityError if already has a supplier.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if self.package_supplier_set:
            raise CardinalityError("Package::Supplier")

        if not validations.validate_pkg_supplier(entity):
            raise SPDXValueError("Package::Supplier")

        self.package_supplier_set = True
        doc.packages[-1].supplier = entity
        return True

    def set_pkg_originator(self, doc, entity):
        """
        Set the package originator, if not already set.
        entity - Organization, Person or NoAssert.
        Raise CardinalityError if already has an originator.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if self.package_originator_set:
            raise CardinalityError("Package::Originator")

        if not validations.validate_pkg_originator(entity):
            raise SPDXValueError("Package::Originator")

        self.package_originator_set = True
        doc.packages[-1].originator = entity
        return True

    def set_pkg_down_location(self, doc, location):
        """
        Set the package download location, if not already set.
        location - A string
        Raise CardinalityError if already defined.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if self.package_down_location_set:
            raise CardinalityError("Package::DownloadLocation")

        self.package_down_location_set = True
        doc.packages[-1].download_location = location
        return True

    def set_pkg_files_analyzed(self, doc, files_analyzed):
        """
        Set the package files analyzed, if not already set.
        Raise SPDXValueError if malformed value, CardinalityError if
        already defined.
        """
        self.assert_package_exists()
        if self.package_files_analyzed_set:
            raise CardinalityError("Package::FilesAnalyzed")

        if files_analyzed is None:
            return True

        if not validations.validate_pkg_files_analyzed(files_analyzed):
            raise SPDXValueError("Package::FilesAnalyzed")

        self.package_files_analyzed_set = True
        if isinstance(files_analyzed, str):
            files_analyzed = files_analyzed.lower() == "true"
        doc.packages[-1].files_analyzed = files_analyzed
        # convert to boolean;
        # validate_pkg_files_analyzed already checked if
        # files_analyzed is in ['True', 'true', 'False', 'false']
        return True

    def set_pkg_home(self, doc, location):
        """Set the package homepage location if not already set.
        location - A string or None or NoAssert.
        Raise CardinalityError if already defined.
        Raise OrderError if no package previously defined.
        Raise SPDXValueError if location has incorrect value.
        """
        self.assert_package_exists()
        if self.package_home_set:
            raise CardinalityError("Package::HomePage")

        if not validations.validate_pkg_homepage(location):
            raise SPDXValueError("Package::HomePage")

        self.package_home_set = True
        doc.packages[-1].homepage = location
        return True

    def set_pkg_verif_code(self, doc, code):
        """
        Set the package verification code, if not already set.
        code - A string.
        Raise CardinalityError if already defined.
        Raise OrderError if no package previously defined.
        Raise Value error if doesn't match verifcode form
        """
        self.assert_package_exists()
        if self.package_verif_set:
            raise CardinalityError("Package::VerificationCode")

        match = self.VERIF_CODE_REGEX.match(code)
        if not match:
            raise SPDXValueError("Package::VerificationCode")

        self.package_verif_set = True
        doc.packages[-1].verif_code = match.group(self.VERIF_CODE_CODE_GRP)

        if match.group(self.VERIF_CODE_EXC_FILES_GRP) is not None:
            doc.packages[-1].verif_exc_files = match.group(
                self.VERIF_CODE_EXC_FILES_GRP
            ).split(",")
        return True

    def set_pkg_checksum(self, doc, checksum):
        """
        Set the package checksum, if not already set.
        checksum - A string
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        self.package_chk_sum_set = True
        doc.packages[-1].set_checksum(Checksum.checksum_from_string(checksum))
        return True

    def set_pkg_source_info(self, doc, text):
        """
        Set the package's source information, if not already set.
        text - Free form text.
        Raise CardinalityError if already defined.
        Raise OrderError if no package previously defined.
        SPDXValueError if text is not free form text or single line of text.
        """
        self.assert_package_exists()
        if self.package_source_info_set:
            raise CardinalityError("Package::SourceInfo")

        if not validations.validate_pkg_src_info(text):
            raise SPDXValueError("Package::SourceInfo")

        self.package_source_info_set = True
        doc.packages[-1].source_info = str_from_text(text)
        return True

    def set_pkg_licenses_concluded(self, doc, licenses):
        """
        Set the package's concluded licenses.
        licenses - License info.
        Raise CardinalityError if already defined.
        Raise OrderError if no package previously defined.
        Raise SPDXValueError if data malformed.
        """
        self.assert_package_exists()
        if self.package_conc_lics_set:
            raise CardinalityError("Package::ConcludedLicenses")

        if not validations.validate_lics_conc(licenses, optional=True):
            raise SPDXValueError("Package::ConcludedLicenses")

        self.package_conc_lics_set = True
        doc.packages[-1].conc_lics = licenses
        return True

    def set_pkg_license_from_file(self, doc, lic):
        """
        Add a license from a file to the package.
        Raise SPDXValueError if data malformed.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if not validations.validate_lics_from_file(lic, optional=True):
            raise SPDXValueError("Package::LicensesFromFile")

        doc.packages[-1].licenses_from_files.append(lic)
        return True

    def set_pkg_license_declared(self, doc, lic):
        """
        Set the package's declared license.
        Raise SPDXValueError if data malformed.
        Raise OrderError if no package previously defined.
        Raise CardinalityError if already set.
        """
        self.assert_package_exists()
        if self.package_license_declared_set:
            raise CardinalityError("Package::LicenseDeclared")

        if not validations.validate_lics_conc(lic, optional=True):
            raise SPDXValueError("Package::LicenseDeclared")

        self.package_license_declared_set = True
        doc.packages[-1].license_declared = lic
        return True

    def set_pkg_license_comment(self, doc, text):
        """
        Set the package's license comment.
        Raise OrderError if no package previously defined.
        Raise CardinalityError if already set.
        Raise SPDXValueError if text is not free form text or single line of text.
        """
        self.assert_package_exists()
        if self.package_license_comment_set:
            raise CardinalityError("Package::LicenseComment")

        if not validations.validate_pkg_lics_comment(text):
            raise SPDXValueError("Package::LicenseComment")

        self.package_license_comment_set = True
        doc.packages[-1].license_comment = str_from_text(text)
        return True

    def set_pkg_attribution_text(self, doc, text):
        """
        Set the package's attribution text .
        Raise SPDXValueError if text is not free form text or single line of text.
        """
        self.assert_package_exists()
        if not validations.validate_pkg_attribution_text(text):
            raise SPDXValueError("Package::AttributionText")

        doc.packages[-1].attribution_text = str_from_text(text)
        return True

    def set_pkg_cr_text(self, doc, text):
        """
        Set the package's copyright text.
        Raise OrderError if no package previously defined.
        Raise CardinalityError if already set.
        Raise value error if text is not one of [None, NOASSERT, TEXT] or single line of text.
        """
        self.assert_package_exists()
        if self.package_cr_text_set:
            raise CardinalityError("Package::CopyrightText")

        if not validations.validate_pkg_cr_text(text, optional=True):
            raise SPDXValueError("Package::CopyrightText")

        self.package_cr_text_set = True
        if isinstance(text, str):
            doc.packages[-1].cr_text = str_from_text(text)
        else:
            doc.packages[-1].cr_text = text  # None or NoAssert

    def set_pkg_summary(self, doc, text):
        """
        Set the package summary.
        Raise SPDXValueError if text is not free form text or single line of text.
        Raise CardinalityError if summary already set.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if self.package_summary_set:
            raise CardinalityError("Package::Summary")

        if not validations.validate_pkg_summary(text):
            raise SPDXValueError("Package::Summary")

        self.package_summary_set = True
        doc.packages[-1].summary = str_from_text(text)

    def set_pkg_desc(self, doc, text):
        """
        Set the package's description.
        Raise SPDXValueError if text is not free form text or single line of text.
        Raise CardinalityError if description already set.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if self.package_desc_set:
            raise CardinalityError("Package::Description")

        if not validations.validate_pkg_desc(text):
            raise SPDXValueError("Package::Description")

        self.package_desc_set = True
        doc.packages[-1].description = str_from_text(text)

    def set_pkg_comment(self, doc, text):
        """
        Set the package's comment.
        Raise SPDXValueError if text is not free form text or single line of text.
        Raise CardinalityError if comment already set.
        Raise OrderError if no package previously defined.
        """
        self.assert_package_exists()
        if self.package_comment_set:
            raise CardinalityError("Package::Comment")

        if not validations.validate_pkg_comment(text):
            raise SPDXValueError("Package::Comment")

        self.package_comment_set = True
        doc.packages[-1].comment = str_from_text(text)

    def set_pkg_primary_package_purpose(self, doc, purpose):
        """
        Set the package's primary purpose.
        Raise CardinalityError if more than one purpose is set.
        Raise SPDXValueError if purpose is unknown.
        """
        self.assert_package_exists()
        if self.package_primary_purpose_set:
            raise CardinalityError("Package::PrimaryPackagePurpose")

        self.package_primary_purpose_set = True
        purpose = purpose.replace("-", "_")
        for purpose_enum in PackagePurpose:
            if purpose == purpose_enum.name:
                doc.packages[-1].primary_package_purpose = purpose_enum
                return True
        else:
            raise SPDXValueError("Package::PrimaryPackagePurpose")

    def set_pkg_built_date(self, doc, built_date):
        """
        Set the package`s built date.
        Raise CardinalityError if built_date date already set.
        Raise SPDXValueError if built_date is not a date.
        """
        self.assert_package_exists()
        if self.package_built_date_set:
            raise CardinalityError("Package::BuiltDate")

        date = utils.datetime_from_iso_format(built_date)
        if date is None:
            raise SPDXValueError("Package::BuiltDate")

        self.package_built_date_set = True
        doc.packages[-1].built_date = date
        return True

    def set_pkg_release_date(self, doc, release_date):
        """
        Set the package`s release date.
        Raise CardinalityError if release_date date already set.
        Raise SPDXValueError if release_date is not a date.
        """
        self.assert_package_exists()
        if self.package_release_date_set:
            raise CardinalityError("Package::ReleaseDate")

        date = utils.datetime_from_iso_format(release_date)
        if date is None:
            raise SPDXValueError("Package::ReleaseDate")

        self.package_release_date_set = True
        doc.packages[-1].release_date = date
        return True

    def set_pkg_valid_until_date(self, doc, valid_until_date):
        """
        Set the package`s valid_until date.
        Raise CardinalityError if valid_until_date date already set.
        Raise SPDXValueError if valid_until_date is not a date.
        """
        self.assert_package_exists()
        if self.package_valid_until_date_set:
            raise CardinalityError("Package::ValidUntilDate")

        date = utils.datetime_from_iso_format(valid_until_date)
        if date is None:
            raise SPDXValueError("Package::ValidUntilDate")

        self.package_valid_until_date_set = True
        doc.packages[-1].valid_until_date = date
        return True

    def set_pkg_ext_ref_category(self, doc, category):
        """
        Set the `category` attribute of the `ExternalPackageRef` object.
        """
        self.assert_package_exists()
        if not validations.validate_pkg_ext_ref_category(category):
            raise SPDXValueError("ExternalRef::Category")

        if (
            len(doc.packages[-1].pkg_ext_refs)
            and doc.packages[-1].pkg_ext_refs[-1].category is None
        ):
            doc.packages[-1].pkg_ext_refs[-1].category = category
        else:
            doc.packages[-1].add_pkg_ext_refs(
                package.ExternalPackageRef(category=category)
            )

    def set_pkg_ext_ref_type(self, doc, pkg_ext_ref_type):
        """
        Set the `pkg_ext_ref_type` attribute of the `ExternalPackageRef` object.
        """
        self.assert_package_exists()
        if not validations.validate_pkg_ext_ref_type(pkg_ext_ref_type):
            raise SPDXValueError("ExternalRef::Type")

        if (
            len(doc.packages[-1].pkg_ext_refs)
            and doc.packages[-1].pkg_ext_refs[-1].pkg_ext_ref_type is None
        ):
            doc.packages[-1].pkg_ext_refs[-1].pkg_ext_ref_type = pkg_ext_ref_type
        else:
            doc.packages[-1].add_pkg_ext_refs(
                package.ExternalPackageRef(pkg_ext_ref_type=pkg_ext_ref_type)
            )

    def set_pkg_ext_ref_locator(self, doc, locator):
        """
        Set the `locator` attribute of the `ExternalPackageRef` object.
        """
        self.assert_package_exists()
        if (
            len(doc.packages[-1].pkg_ext_refs)
            and doc.packages[-1].pkg_ext_refs[-1].locator is None
        ):
            doc.packages[-1].pkg_ext_refs[-1].locator = locator
        else:
            doc.packages[-1].add_pkg_ext_refs(package.ExternalPackageRef(locator=locator))

    def add_pkg_ext_ref_comment(self, doc, comment):
        """
        Set the `comment` attribute of the `ExternalPackageRef` object.
        """
        self.assert_package_exists()
        if not len(doc.packages[-1].pkg_ext_refs):
            raise OrderError("Package::ExternalRef")

        if not validations.validate_pkg_ext_ref_comment(comment):
            raise SPDXValueError("ExternalRef::Comment")

        doc.packages[-1].pkg_ext_refs[-1].comment = str_from_text(comment)

    def add_pkg_ext_refs(self, doc, category, pkg_ext_ref_type, locator):
        self.set_pkg_ext_ref_category(doc, category)
        self.set_pkg_ext_ref_type(doc, pkg_ext_ref_type)
        self.set_pkg_ext_ref_locator(doc, locator)

    def assert_package_exists(self):
        if not self.package_set:
            raise OrderError("Package")


class FileBuilder(object):
    def __init__(self):
        # FIXME: this state does not make sense
        self.reset_file_stat()

    def set_file_name(self, doc, name):
        doc.files.append(file.File(name))
        # A file name marks the start of a new file instance.
        # The builder must be reset
        # FIXME: this state does not make sense
        self.reset_file_stat()
        return True

    def set_file_spdx_id(self, doc, spdx_id):
        """
        Set the file SPDX Identifier.
        Raise OrderError if no package or no file defined.
        Raise SPDXValueError if malformed value.
        Raise CardinalityError if more than one spdx_id set.
        """
        if not self.has_file(doc):
            raise OrderError("File::SPDXID")

        if self.file_spdx_id_set:
            raise CardinalityError("File::SPDXID")

        if not validations.validate_file_spdx_id(spdx_id):
            raise SPDXValueError("File::SPDXID")

        self.file_spdx_id_set = True
        self.file(doc).spdx_id = spdx_id
        return True

    def set_file_comment(self, doc, text):
        """
        Raise OrderError if no package or no file defined.
        Raise CardinalityError if more than one comment set.
        Raise SPDXValueError if text is not free form text or single line of text.
        """
        if not self.has_file(doc):
            raise OrderError("File::Comment")

        if self.file_comment_set:
            raise CardinalityError("File::Comment")

        if not validations.validate_file_comment(text):
            raise SPDXValueError("File::Comment")

        self.file_comment_set = True
        self.file(doc).comment = str_from_text(text)
        return True

    def set_file_attribution_text(self, doc, text):
        """
        Set the file's attribution text .
        Raise OrderError if no package or no file defined.
        Raise SPDXValueError if text is not free form text or single line of text.
        """
        if not self.has_file(doc):
            raise OrderError("File::AttributionText")

        if not validations.validate_file_attribution_text(text):
            raise SPDXValueError("File::AttributionText")

        self.file(doc).attribution_text = str_from_text(text)
        return True

    def set_file_type(self, doc, type_value):
        """
        Raise OrderError if no package or file defined.
        Raise CardinalityError if more than one type set.
        Raise SPDXValueError if type is unknown.
        """
        if not self.has_file(doc):
            raise OrderError("File::FileType")

        if type_value not in file.FileType.__members__:
            raise SPDXValueError("File:FileType")

        file_type = file.FileType[type_value]

        spdx_file = self.file(doc)
        if file_type in spdx_file.file_types:
            raise CardinalityError("File::FileType")

        spdx_file.file_types.append(file_type)

    def set_file_checksum(self, doc: Document, checksum: str):
        """
        Raise OrderError if no file defined.
        """
        if self.has_file(doc):
            new_checksum = Checksum.checksum_from_string(checksum)
            self.file(doc).set_checksum(new_checksum)
        else:
            raise OrderError("File::CheckSum")
        return True

    def set_concluded_license(self, doc, lic):
        """
        Raise OrderError if no package or file defined.
        Raise CardinalityError if already set.
        Raise SPDXValueError if malformed.
        """
        if not self.has_file(doc):
            raise OrderError("File::ConcludedLicense")

        if self.file_conc_lics_set:
            raise CardinalityError("File::ConcludedLicense")

        if not validations.validate_lics_conc(lic, optional=True):
            raise SPDXValueError("File::ConcludedLicense")

        self.file_conc_lics_set = True
        self.file(doc).conc_lics = lic
        return True

    def set_file_license_in_file(self, doc, lic):
        """
        Raise OrderError if no package or file defined.
        Raise SPDXValueError if malformed value.
        """
        if not self.has_file(doc):
            raise OrderError("File::LicenseInFile")

        if not validations.validate_file_lics_in_file(lic):
            raise SPDXValueError("File::LicenseInFile")

        self.file(doc).add_lics(lic)
        return True

    def set_file_license_comment(self, doc, text):
        """
        Raise OrderError if no package or file defined.
        Raise SPDXValueError if text is not free form text or single line of text.
        Raise CardinalityError if more than one per file.
        """
        if not self.has_file(doc):
            raise OrderError("File::LicenseComment")

        if self.file_license_comment_set:
            raise CardinalityError("File::LicenseComment")

        if not validations.validate_file_lics_comment(text):
            raise SPDXValueError("File::LicenseComment")

        self.file_license_comment_set = True
        self.file(doc).license_comment = str_from_text(text)

    def set_file_copyright(self, doc, text):
        """
        Raise OrderError if no package or file defined.
        Raise SPDXValueError if not free form text or NONE or NO_ASSERT or single line of text.
        Raise CardinalityError if more than one.
        """
        if not self.has_file(doc):
            raise OrderError("File::CopyRight")

        if self.file_copytext_set:
            raise CardinalityError("File::CopyRight")

        if not validations.validate_file_cpyright(text, optional=True):
            raise SPDXValueError("File::CopyRight")

        self.file_copytext_set = True
        if isinstance(text, str):
            self.file(doc).copyright = str_from_text(text)
        else:
            self.file(doc).copyright = text  # None or NoAssert
        return True

    def set_file_notice(self, doc, text):
        """
        Raise OrderError if no package or file defined.
        Raise SPDXValueError if not free form text or single line of text.
        Raise CardinalityError if more than one.
        """
        if not self.has_file(doc):
            raise OrderError("File::Notice")

        if self.file_notice_set:
            raise CardinalityError("File::Notice")

        if not validations.validate_file_notice(text):
            raise SPDXValueError("File::Notice")

        self.file_notice_set = True
        self.file(doc).notice = str_from_text(text)

    def add_file_contribution(self, doc, value):
        """
        Raise OrderError if no package or file defined.
        """
        if not self.has_file(doc):
            raise OrderError("File::Contributor")

        self.file(doc).add_contrib(value)

    def add_file_dep(self, doc, value):
        """
        Raise OrderError if no package or file defined.
        """
        if not self.has_file(doc):
            raise OrderError("File::Dependency")

        self.file(doc).add_depend(value)

    def set_file_atrificat_of_project(self, doc, symbol, value):
        """
        Set a file name, uri or home artifact.
        Raise OrderError if no package or file defined.
        """
        if not self.has_file(doc):
            raise OrderError("File::Artifact")

        self.file(doc).add_artifact(symbol, value)

    def file(self, doc):
        """
        Return the last file in the document's file list.
        """
        return doc.files[-1]

    def has_file(self, doc):
        """
        Return true if the document has at least one file.
        """
        return len(doc.files) != 0

    def has_package(self, doc):
        """
        Return true if the document has a package.
        """
        return len(doc.packages) != 0

    def reset_file_stat(self):
        """
        Reset the builder's state to enable building new files.
        """
        # FIXME: this state does not make sense
        self.file_spdx_id_set = False
        self.file_comment_set = False
        self.file_type_set = False
        self.file_chksum_set = False
        self.file_conc_lics_set = False
        self.file_license_comment_set = False
        self.file_notice_set = False
        self.file_copytext_set = False


class LicenseBuilder(object):
    def __init__(self):
        # FIXME: this state does not make sense
        self.reset_extr_lics()

    def extr_lic(self, doc):
        """
        Retrieve last license in extracted license list.
        """
        return doc.extracted_licenses[-1]

    def has_extr_lic(self, doc):
        return len(doc.extracted_licenses) != 0

    def set_lic_id(self, doc, lic_id):
        """
        Add a new extracted license to the document.
        Raise SPDXValueError if data format is incorrect.
        """
        # FIXME: this state does not make sense
        self.reset_extr_lics()
        if not validations.validate_extracted_lic_id(lic_id):
            raise SPDXValueError("ExtractedLicense::id")

        doc.add_extr_lic(license.ExtractedLicense(lic_id))
        return True

    def set_lic_text(self, doc, text):
        """
        Set license extracted text.
        Raise SPDXValueError if text is not free form text or single line of text.
        Raise OrderError if no license ID defined.
        """
        if not self.has_extr_lic(doc):
            raise OrderError("ExtractedLicense::text")

        if self.extr_text_set:
            raise CardinalityError("ExtractedLicense::text")

        if not validations.validate_is_free_form_text_or_str(text):
            raise SPDXValueError("ExtractedLicense::text")

        self.extr_text_set = True
        self.extr_lic(doc).text = str_from_text(text)
        return True

    def set_lic_name(self, doc, name):
        """
        Set license name.
        Raise SPDXValueError if name is not str or utils.NoAssert
        Raise OrderError if no license id defined.
        """
        if not self.has_extr_lic(doc):
            raise OrderError("ExtractedLicense::Name")

        if self.extr_lic_name_set:
            raise CardinalityError("ExtractedLicense::Name")

        if not validations.validate_extr_lic_name(name):
            raise SPDXValueError("ExtractedLicense::Name")

        self.extr_lic_name_set = True
        self.extr_lic(doc).full_name = name
        return True

    def set_lic_comment(self, doc, comment):
        """
        Set license comment.
        Raise SPDXValueError if comment is not free form text or single line of text.
        Raise OrderError if no license ID defined.
        """
        if not self.has_extr_lic(doc):
            raise OrderError("ExtractedLicense::comment")

        if self.extr_lic_comment_set:
            raise CardinalityError("ExtractedLicense::comment")

        if not validations.validate_is_free_form_text_or_str(comment):
            raise SPDXValueError("ExtractedLicense::comment")

        self.extr_lic_comment_set = True
        self.extr_lic(doc).comment = str_from_text(comment)
        return True

    def add_lic_xref(self, doc, ref):
        """
        Add a license cross reference.
        Raise OrderError if no License ID defined.
        """
        if not self.has_extr_lic(doc):
            raise OrderError("ExtractedLicense::CrossRef")

        self.extr_lic(doc).add_xref(ref)
        return True

    def reset_extr_lics(self):
        # FIXME: this state does not make sense
        self.extr_text_set = False
        self.extr_lic_name_set = False
        self.extr_lic_comment_set = False


class SnippetBuilder(object):
    def __init__(self):
        # FIXME: this state does not make sense
        self.reset_snippet()

    def create_snippet(self, doc, spdx_id):
        """
        Create a snippet for the SPDX Document.
        spdx_id - To uniquely identify any element in an SPDX document which
        may be referenced by other elements.
        Raise SPDXValueError if the data is a malformed value.
        """
        self.reset_snippet()
        spdx_id = spdx_id.split("#")[-1]
        if not validations.validate_snippet_spdx_id(spdx_id):
            raise SPDXValueError("Snippet::SnippetSPDXID")

        self.snippet_spdx_id_set = True
        doc.add_snippet(snippet.Snippet(spdx_id=spdx_id))
        return True

    def set_snippet_name(self, doc, name):
        """
        Set name of the snippet.
        Raise OrderError if no snippet previously defined.
        Raise CardinalityError if the name is already set.
        """
        self.assert_snippet_exists()
        if self.snippet_name_set:
            raise CardinalityError("SnippetName")

        self.snippet_name_set = True
        doc.snippet[-1].name = name
        return True

    def set_snippet_comment(self, doc, comment):
        """
        Set general comments about the snippet.
        Raise OrderError if no snippet previously defined.
        Raise SPDXValueError if the data is not free form text or single line of text.
        Raise CardinalityError if comment already set.
        """
        self.assert_snippet_exists()
        if self.snippet_comment_set:
            raise CardinalityError("Snippet::SnippetComment")

        if not validations.validate_snip_comment(comment):
            raise SPDXValueError("Snippet::SnippetComment")

        self.snippet_comment_set = True
        doc.snippet[-1].comment = str_from_text(comment)
        return True

    def set_snippet_attribution_text(self, doc, text):
        """
        Set the snippet's attribution text .
        Raise SPDXValueError if text is not free form text or single line of text.
        """
        self.assert_snippet_exists()
        if not validations.validate_snippet_attribution_text(text):
            raise SPDXValueError("Snippet::AttributionText")

        doc.snippet[-1].attribution_text = str_from_text(text)
        return True

    def set_snippet_copyright(self, doc, text):
        """Set the snippet's copyright text.
        Raise OrderError if no snippet previously defined.
        Raise CardinalityError if already set.
        Raise SPDXValueError if text is not one of [None, NOASSERT, TEXT] or single line of text.
        """
        self.assert_snippet_exists()
        if self.snippet_copyright_set:
            raise CardinalityError("Snippet::SnippetCopyrightText")

        if not validations.validate_snippet_copyright(text, optional=True):
            raise SPDXValueError("Snippet::SnippetCopyrightText")

        self.snippet_copyright_set = True
        if isinstance(text, str):
            doc.snippet[-1].copyright = str_from_text(text)
        else:
            doc.snippet[-1].copyright = text  # None or NoAssert
        return True

    def set_snippet_lic_comment(self, doc, text):
        """
        Set the snippet's license comment.
        Raise OrderError if no snippet previously defined.
        Raise CardinalityError if already set.
        Raise SPDXValueError if the data is not free form text or single line of text.
        """
        self.assert_snippet_exists()
        if self.snippet_lic_comment_set:
            raise CardinalityError("Snippet::SnippetLicenseComments")

        if not validations.validate_snip_lic_comment(text):
            raise SPDXValueError("Snippet::SnippetLicenseComments")

        self.snippet_lic_comment_set = True
        doc.snippet[-1].license_comment = str_from_text(text)
        return True

    def set_snip_from_file_spdxid(self, doc, snip_from_file_spdxid):
        """
        Set the snippet's 'Snippet from File SPDX Identifier'.
        Raise OrderError if no snippet previously defined.
        Raise CardinalityError if already set.
        Raise SPDXValueError if the data is a malformed value.
        """
        self.assert_snippet_exists()
        snip_from_file_spdxid = snip_from_file_spdxid.split("#")[-1]
        if self.snip_file_spdxid_set:
            raise CardinalityError("Snippet::SnippetFromFileSPDXID")

        if not validations.validate_snip_file_spdxid(snip_from_file_spdxid):
            raise SPDXValueError("Snippet::SnippetFromFileSPDXID")

        self.snip_file_spdxid_set = True
        doc.snippet[-1].snip_from_file_spdxid = snip_from_file_spdxid
        return True

    def set_snip_concluded_license(self, doc, conc_lics):
        """
        Raise OrderError if no snippet previously defined.
        Raise CardinalityError if already set.
        Raise SPDXValueError if the data is a malformed value.
        """
        self.assert_snippet_exists()
        if self.snippet_conc_lics_set:
            raise CardinalityError("Snippet::SnippetLicenseConcluded")

        if not validations.validate_lics_conc(conc_lics, optional=True):
            raise SPDXValueError("Snippet::SnippetLicenseConcluded")

        self.snippet_conc_lics_set = True
        doc.snippet[-1].conc_lics = conc_lics
        return True

    def set_snippet_lics_info(self, doc, lics_info):
        """
        Raise OrderError if no snippet previously defined.
        Raise SPDXValueError if the data is a malformed value.
        """
        self.assert_snippet_exists()
        if not validations.validate_snip_lics_info(lics_info, optional=True):
            raise SPDXValueError("Snippet::LicenseInfoInSnippet")

        doc.snippet[-1].add_lics(lics_info)
        return True

    def set_snippet_byte_range(self, doc, parsed):
        """
        Raise OrderError if no snippet previously defined.
        Raise SPDXValueError if the data is malformed.
        """
        self.assert_snippet_exists()
        range_re = re.compile(r"^(\d+):(\d+)$", re.UNICODE)
        if not range_re.match(parsed):
            raise SPDXValueError("Snippet:ByteRange")
        startpoint = int(range_re.match(parsed).group(1))
        endpoint = int(range_re.match(parsed).group(2))
        if startpoint <= endpoint:
            doc.snippet[-1].byte_range = (startpoint, endpoint)
        else:
            raise SPDXValueError("Snippet::ByteRange")

    def set_snippet_line_range(self, doc, parsed):
        """
        Raise OrderError if no snippet previously defined.
        Raise SPDXValueError if the data is malformed.
        """
        self.assert_snippet_exists()
        range_re = re.compile(r"^(\d+):(\d+)$", re.UNICODE)
        if not range_re.match(parsed):
            raise SPDXValueError("Snippet:LineRange")
        startpoint = int(range_re.match(parsed).group(1))
        endpoint = int(range_re.match(parsed).group(2))
        if startpoint <= endpoint:
            doc.snippet[-1].line_range = (startpoint, endpoint)
        else:
            raise SPDXValueError("Snippet::LineRange")

    def reset_snippet(self):
        # FIXME: this state does not make sense
        self.snippet_spdx_id_set = False
        self.snippet_name_set = False
        self.snippet_comment_set = False
        self.snippet_copyright_set = False
        self.snippet_lic_comment_set = False
        self.snip_file_spdxid_set = False
        self.snippet_conc_lics_set = False

    def assert_snippet_exists(self):
        if not self.snippet_spdx_id_set:
            raise OrderError("Snippet")


class Builder(
    DocBuilder,
    CreationInfoBuilder,
    EntityBuilder,
    ReviewBuilder,
    PackageBuilder,
    FileBuilder,
    LicenseBuilder,
    SnippetBuilder,
    ExternalDocumentRefBuilder,
    AnnotationBuilder,
    RelationshipBuilder,
):
    """
    SPDX document builder.
    """

    def __init__(self):
        super(Builder, self).__init__()
        # FIXME: this state does not make sense
        self.reset()
        self.current_package: Dict = dict()
        self.current_file: Dict = dict()

    def set_current_package_name(self, name: str) -> None:
        self.current_package["name"] = name

    def set_current_file_name(self, name: str) -> None:
        self.current_file["name"] = name

    def set_current_file_id(self, spdx_id: str) -> None:
        self.current_file["spdx_id"] = spdx_id

    def set_current_package_id(self, spdx_id: str) -> None:
        self.current_package["spdx_id"] = spdx_id

    def current_package_has_name(self) -> bool:
        return bool(("name" in self.current_package) and (self.current_package["name"]))

    def current_file_has_name(self) -> bool:
        return bool(("name" in self.current_file) and (self.current_file["name"]))

    def current_package_has_id(self) -> bool:
        return bool("spdx_id" in self.current_package) and (self.current_package["spdx_id"])

    def current_file_has_id(self) -> bool:
        return bool("spdx_id" in self.current_file) and (self.current_file["spdx_id"])

    def has_current_package(self) -> bool:
        return bool(self.current_package)

    def reset(self):
        """
        Reset builder's state for building new documents.
        Must be called between usage with different documents.
        """
        # FIXME: this state does not make sense
        self.reset_creation_info()
        self.reset_document()
        self.reset_package()
        self.reset_file_stat()
        self.reset_reviews()
        self.reset_annotations()
        self.reset_extr_lics()
        self.reset_snippet()
        self.reset_relationship()



# Copyright (c) 2022 spdx contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from spdx.checksum import ChecksumAlgorithm, Checksum


@pytest.mark.parametrize("algorithm,expected",
                         [("SHA1", "checksumAlgorithm_sha1"), ("SHA224", "checksumAlgorithm_sha224"),
                          ("SHA3_256", "checksumAlgorithm_sha3_256"), ("BLAKE2B_256", "checksumAlgorithm_blake2b256"),
                          ("MD5", "checksumAlgorithm_md5")])
def test_checksum_to_rdf(algorithm, expected):
    test_algorithm = ChecksumAlgorithm[algorithm]
    rdf_algorithm = test_algorithm.algorithm_to_rdf_representation()

    assert rdf_algorithm == expected


@pytest.mark.parametrize("expected,rdf_algorithm",
                         [(ChecksumAlgorithm.SHA1, "checksumAlgorithm_sha1"),
                          (ChecksumAlgorithm.SHA224, "checksumAlgorithm_sha224"),
                          (ChecksumAlgorithm.SHA3_256, "checksumAlgorithm_sha3_256"),
                          (ChecksumAlgorithm.BLAKE2B_256, "checksumAlgorithm_blake2b256"),
                          (ChecksumAlgorithm.MD5, "checksumAlgorithm_md5")])
def test_checksum_from_rdf(rdf_algorithm, expected):
    algorithm = ChecksumAlgorithm.checksum_from_rdf(rdf_algorithm)

    assert algorithm == expected


@pytest.mark.parametrize("rdf_algorithm",
                         ["_checksumAlgorithm_sha1", "checksumAlgorithm_sha_224", "checksumAlgorithm_sha3256",
                          "checksumAlgorithm_blake2b 256", "checksumAlgorithm_blake2b-256",
                          "checksumAlgorithm_bblake2b 256"])
def test_checksum_from_wrong_rdf(rdf_algorithm):
    with pytest.raises(ValueError) as error:
        ChecksumAlgorithm.checksum_from_rdf(rdf_algorithm)

    assert str(error.value).startswith("Invalid algorithm for checksum")


CHECKSUM_VALUE = "123abc"


@pytest.mark.parametrize("checksum_string,expected",
                         [("SHA1: " + CHECKSUM_VALUE, Checksum(ChecksumAlgorithm.SHA1, CHECKSUM_VALUE)),
                          ("SHA3-256: " + CHECKSUM_VALUE, Checksum(ChecksumAlgorithm.SHA3_256, CHECKSUM_VALUE)),
                          ("ADLER32: " + CHECKSUM_VALUE, Checksum(ChecksumAlgorithm.ADLER32, CHECKSUM_VALUE)),
                          ("BLAKE3: " + CHECKSUM_VALUE, Checksum(ChecksumAlgorithm.BLAKE3, CHECKSUM_VALUE)),
                          ("BLAKE2b-256: " + CHECKSUM_VALUE, Checksum(ChecksumAlgorithm.BLAKE2B_256, CHECKSUM_VALUE)),
                          ("MD5: " + CHECKSUM_VALUE, Checksum(ChecksumAlgorithm.MD5, CHECKSUM_VALUE))])
def test_checksum_from_string(checksum_string: str, expected: Checksum):
    checksum: Checksum = Checksum.checksum_from_string(checksum_string)
    assert checksum == expected


@pytest.mark.parametrize("checksum_string", ["SHA1: ABC", "SHA1000: abc"])
def test_wrong_checksum_from_string(checksum_string: str):
    with pytest.raises(ValueError, match=f"Invalid checksum: {checksum_string}"):
        Checksum.checksum_from_string(checksum_string)


@pytest.mark.parametrize("checksum, expected",
                         [(Checksum(ChecksumAlgorithm.SHA1, CHECKSUM_VALUE), "SHA1: " + CHECKSUM_VALUE),
                          (Checksum(ChecksumAlgorithm.SHA3_256, CHECKSUM_VALUE), "SHA3-256: " + CHECKSUM_VALUE),
                          (Checksum(ChecksumAlgorithm.ADLER32, CHECKSUM_VALUE), "ADLER32: " + CHECKSUM_VALUE),
                          (Checksum(ChecksumAlgorithm.BLAKE3, CHECKSUM_VALUE), "BLAKE3: " + CHECKSUM_VALUE),
                          (Checksum(ChecksumAlgorithm.BLAKE2B_256, CHECKSUM_VALUE), "BLAKE2b-256: " + CHECKSUM_VALUE),
                          (Checksum(ChecksumAlgorithm.MD5, CHECKSUM_VALUE), "MD5: " + CHECKSUM_VALUE)])
def test_checksum_to_tv(checksum: Checksum, expected: str):
    checksum_string: str = checksum.to_tv()
    assert checksum_string == expected



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import rdflib
import uritools

from spdx import creationinfo
from spdx import license
from spdx import utils


def validate_is_free_form_text_or_str(value, optional=False) -> bool:
    if value is None:
        return optional
    if not isinstance(value, str):
        return False
    if "\n" in value:
        TEXT_RE = re.compile(r"<text>(.|\n)*</text>", re.UNICODE)
        match = TEXT_RE.match(value)
        return match is not None
    return True


def validate_tool_name(value, optional=False):
    striped_value = value.strip()
    if optional:
        return len(striped_value) == 0
    else:
        return not (len(striped_value) == 0)


def validate_person_name(value, optional=False):
    return validate_tool_name(value, optional)


def validate_org_name(value, optional=False):
    return validate_tool_name(value, optional)


def validate_data_lics(value):
    return value == "CC0-1.0"


def validate_doc_name(value, optional=False):
    return validate_tool_name(value, optional)


def validate_pkg_supplier(value, optional=False):
    if optional and value is None:
        return True
    return isinstance(
        value, (utils.NoAssert, creationinfo.Person, creationinfo.Organization)
    )


def validate_pkg_originator(value, optional=False):
    return validate_pkg_supplier(value, optional)


def validate_pkg_homepage(value, optional=False):
    if value is None:
        return optional

    return isinstance(value, (str, utils.NoAssert, utils.SPDXNone))


def validate_pkg_cr_text(value, optional=False):
    if isinstance(value, (utils.NoAssert, utils.SPDXNone)):
        return True
    elif validate_is_free_form_text_or_str(value, optional):
        return True
    elif value is None:
        return optional
    else:
        return False


def validate_pkg_summary(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_pkg_desc(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_pkg_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_pkg_attribution_text(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_file_attribution_text(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_snippet_attribution_text(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_pkg_ext_ref_category(value, optional=False):
    # PACKAGE_MANAGER is used in the json schema for 2.2. For now, we simply allow both versions
    return value.upper() in ["SECURITY", "OTHER", "PACKAGE-MANAGER", "PACKAGE_MANAGER"]


def validate_pkg_ext_ref_type(value, optional=False):
    return re.match(r"^\S+$", value) is not None


def validate_pkg_ext_ref_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_doc_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_doc_spdx_id(value, optional=False):
    if value is None:
        return optional

    return value.endswith("#SPDXRef-DOCUMENT")


def validate_doc_namespace(value, optional=False):
    if value is None:
        return optional

    return uritools.isabsuri(value) and ("#" not in value)


def validate_creator(value, optional=False):
    if value is None:
        return optional
    else:
        return isinstance(value, creationinfo.Creator)


def validate_creation_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_reviewer(value, optional=False):
    return validate_creator(value, optional)


def validate_review_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_annotator(value, optional=False):
    return validate_creator(value, optional)


def validate_annotation_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_annotation_type(value, optional=False):
    value = value.strip()
    return value == "REVIEW" or value == "OTHER"


def validate_relationship_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_pkg_spdx_id(value, optional=False):
    value = value.split("#")[-1]
    TEXT_RE = re.compile(r"SPDXRef-([A-Za-z0-9\.\-]+)", re.UNICODE)
    if value is None:
        return optional
    else:
        return TEXT_RE.match(value) is not None


def validate_pkg_files_analyzed(value, optional=False):
    if value in ["True", "true", "False", "false", True, False]:
        return True
    else:
        return optional


def validate_pkg_src_info(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_pkg_lics_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_file_spdx_id(value, optional=False):
    value = value.split("#")[-1]
    TEXT_RE = re.compile(r"SPDXRef-([A-Za-z0-9.\-]+)", re.UNICODE)
    if value is None:
        return optional
    else:
        return TEXT_RE.match(value) is not None


def validate_file_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_file_lics_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_file_cpyright(value, optional=False):
    if isinstance(value, (utils.NoAssert, utils.SPDXNone)):
        return True
    return validate_is_free_form_text_or_str(value, optional)


def validate_lics_from_file(value, optional=False):
    if value is None:
        return optional
    return isinstance(value, (license.License, utils.SPDXNone, utils.NoAssert))


def validate_file_notice(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_lics_conc(value, optional=False):
    if value is None:
        return optional
    return isinstance(value, (utils.NoAssert, utils.SPDXNone, license.License))


def validate_file_lics_in_file(value, optional=False):
    if value is None:
        return optional
    return isinstance(value, (utils.NoAssert, utils.SPDXNone, license.License))


def validate_extracted_lic_id(value, optional=False):
    if value is None:
        return optional
    else:
        return value.startswith("LicenseRef-")


def validate_extr_lic_name(value, optional=False):
    if value is None:
        return optional
    else:
        return isinstance(value, (str, utils.NoAssert, rdflib.Literal))


def validate_snippet_spdx_id(value, optional=False):
    if value is None:
        return optional
    value = value.split("#")[-1]
    return re.match(r"^SPDXRef[A-Za-z0-9.\-]+$", value) is not None


def validate_snip_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_snippet_copyright(value, optional=False):
    if validate_is_free_form_text_or_str(value, optional):
        return True
    elif isinstance(value, (utils.NoAssert, utils.SPDXNone)):
        return True
    elif value is None:
        return optional
    else:
        return False


def validate_snip_lic_comment(value, optional=False):
    return validate_is_free_form_text_or_str(value, optional)


def validate_snip_file_spdxid(value, optional=False):
    if value is None:
        return optional
    return (
        re.match(r"(DocumentRef[A-Za-z0-9.\-]+:){0,1}SPDXRef[A-Za-z0-9.\-]+", value)
        is not None
    )


def validate_snip_lics_info(value, optional=False):
    if value is None:
        return optional
    return isinstance(value, (utils.NoAssert, utils.SPDXNone, license.License))



import os
from typing import Optional
from unittest import TestCase

import pytest
from rdflib import URIRef

from spdx.document import Document
from spdx.license import License
from spdx.package import Package, ExternalPackageRef
from spdx.parsers.loggers import StandardLogger
from spdx.parsers.parse_anything import parse_file
from spdx.parsers.rdf import Parser
from spdx.parsers.rdfbuilders import Builder
from spdx.utils import NoAssert
from spdx.writers.rdf import Writer


@pytest.fixture
def temporary_file_path() -> str:
    temporary_file_path = "temp_accept_provided_doc_node.rdf.xml"
    yield temporary_file_path
    os.remove(temporary_file_path)


# This test is really clunky since it's hard to isolate features of the rdf writer to test. Should be improved when
# that part is refactored.
def test_accept_provided_doc_node(temporary_file_path) -> None:
    doc_node = URIRef("http://www.spdx.org/tools#SPDXRef-DOCUMENT")
    document: Document = minimal_document_with_package()

    with open(temporary_file_path, "wb") as out:
        writer = Writer(document, out)
        writer.write(doc_node)
    parser = Parser(Builder(), StandardLogger())
    with open(temporary_file_path, "r") as file:
        parsed_document: Document = parser.parse(file)[0]

    # These properties are set automatically if no doc_node is provided. Instead, we provided an empty one
    assert parsed_document.data_license is None
    assert parsed_document.namespace is None
    assert parsed_document.spdx_id is None


def test_external_package_references(temporary_file_path) -> None:
    document: Document = minimal_document_with_package()
    package: Package = document.packages[0]
    first_ref = ExternalPackageRef(category="PACKAGE-MANAGER")
    second_ref = ExternalPackageRef(category="SECURITY")
    package.add_pkg_ext_refs(first_ref)
    package.add_pkg_ext_refs(second_ref)

    # Not using write_anything here because we don't want to deal with validation
    with open(temporary_file_path, "wb") as out:
        writer = Writer(document, out)
        writer.write()

    parsed_document = parse_file(temporary_file_path)[0]
    parsed_package: Package = parsed_document.packages[0]

    assert len(parsed_package.pkg_ext_refs) is 2
    parsed_reference_categories = list(map(lambda x: x.category, parsed_package.pkg_ext_refs))
    assert first_ref.category in parsed_reference_categories
    assert second_ref.category in parsed_reference_categories


# This test is really clunky since it's hard to isolate features of the rdf writer to test. Should be improved when
# that part is refactored.
def test_multiple_packages_in_one_document(temporary_file_path) -> None:
    doc_node = URIRef("http://www.spdx.org/tools#SPDXRef-DOCUMENT")
    document = Document()
    document.creation_info.set_created_now()
    package = Package()
    package.spdx_id = "SPDXRef-Package"
    package.version = "2.1"
    document.add_package(package)
    package2 = Package()
    package2.spdx_id = "SPDXRef-Another-Package"
    package2.version = "2.3"
    document.add_package(package2)

    with open(temporary_file_path, "wb") as out:
        writer = Writer(document, out)
        writer.write(doc_node)
    parser = Parser(Builder(), StandardLogger())
    with open(temporary_file_path, "r") as file:
        parsed_document: Document = parser.parse(file)[0]

    assert len(parsed_document.packages) == 2
    first_package = get_package_by_spdx_id("SPDXRef-Package", document)
    assert first_package.version == "2.1"
    second_package = get_package_by_spdx_id("SPDXRef-Another-Package", document)
    assert second_package.version == "2.3"


def minimal_document_with_package() -> Document:
    document = Document(data_license=License.from_identifier('CC0-1.0'))
    document.creation_info.set_created_now()
    package: Package = minimal_package()
    document.add_package(package)
    return document


def minimal_package() -> Package:
    package = Package()
    package.spdx_id = "SPDXRef-Package"
    package.conc_lics = NoAssert()
    package.license_declared = NoAssert()
    package.add_lics_from_file(NoAssert())
    return package


def get_package_by_spdx_id(package_spdx_id: str, document: Document) -> Optional[Package]:
    for package in document.packages:
        if package.spdx_id == package_spdx_id:
            return package
    return None



# Copyright (c) Xavier Figueroa
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Union

from spdx.document import Document
from spdx.parsers import rdfbuilders
from spdx.parsers import tagvaluebuilders
from spdx.parsers import validations
from spdx.checksum import Checksum, ChecksumAlgorithm
from spdx.parsers.builderexceptions import SPDXValueError
from spdx.parsers.builderexceptions import CardinalityError
from spdx.parsers.builderexceptions import OrderError


class CreationInfoBuilder(rdfbuilders.CreationInfoBuilder):
    def __init__(self):
        super(CreationInfoBuilder, self).__init__()


class ExternalDocumentRefsBuilder(rdfbuilders.ExternalDocumentRefBuilder):
    def __init__(self):
        super(ExternalDocumentRefsBuilder, self).__init__()


class EntityBuilder(rdfbuilders.EntityBuilder):
    def __init__(self):
        super(EntityBuilder, self).__init__()


class SnippetBuilder(rdfbuilders.SnippetBuilder):
    def __init__(self):
        super(SnippetBuilder, self).__init__()


class ReviewBuilder(rdfbuilders.ReviewBuilder):
    def __init__(self):
        super(ReviewBuilder, self).__init__()


class PackageBuilder(rdfbuilders.PackageBuilder):
    def __init__(self):
        super(PackageBuilder, self).__init__()


class DocBuilder(tagvaluebuilders.DocBuilder):
    def __init__(self):
        super(DocBuilder, self).__init__()

    def set_doc_spdx_id(self, doc, doc_spdx_id_line):
        """
        Set the document SPDX Identifier.
        Raise SPDXValueError if malformed value, CardinalityError
        if already defined.
        """
        if not self.doc_spdx_id_set:
            if (
                doc_spdx_id_line == "SPDXRef-DOCUMENT"
                or validations.validate_doc_spdx_id(doc_spdx_id_line)
            ):
                doc.spdx_id = doc_spdx_id_line
                self.doc_spdx_id_set = True
                return True
            else:
                raise SPDXValueError("Document::SPDXID")
        else:
            raise CardinalityError("Document::SPDXID")

    def set_doc_comment(self, doc, comment):
        """
        Set document comment.
        Raise CardinalityError if comment already set.
        """
        if not self.doc_comment_set:
            self.doc_comment_set = True
            doc.comment = comment
        else:
            raise CardinalityError("Document::Comment")
    
    def set_doc_namespace(self, doc, namespace):
        """
        Set the document namespace.
        Raise SPDXValueError if malformed value.
        Raise CardinalityError if already defined.
        """
        if not self.doc_namespace_set:
            self.doc_namespace_set = True
            if validations.validate_doc_namespace(namespace):
                doc.namespace = namespace
                return True
            else:
                raise SPDXValueError("Document::Namespace")
        else:
            raise CardinalityError("Document::Comment")


class LicenseBuilder(tagvaluebuilders.LicenseBuilder):
    def __init__(self):
        super(LicenseBuilder, self).__init__()

    def set_lic_name(self, doc, name):
        """
        Set license name.
        Raise SPDXValueError if name is not str or utils.NoAssert
        Raise CardinalityError if it is already set
        Raise OrderError if no license id defined.
        """
        if self.has_extr_lic(doc):
            if not self.extr_lic_name_set:
                self.extr_lic_name_set = True
                if validations.validate_extr_lic_name(name, True):
                    self.extr_lic(doc).full_name = name
                    return True
                else:
                    raise SPDXValueError("ExtractedLicense::Name")
            else:
                raise CardinalityError("ExtractedLicense::Name")
        else:
            raise OrderError("ExtractedLicense::Name")

    def set_lic_text(self, doc, text):
        """
        Set license name.
        Raise CardinalityError if it is already set.
        Raise OrderError if no license id defined.
        """
        if self.has_extr_lic(doc):
            if not self.extr_text_set:
                self.extr_text_set = True
                self.extr_lic(doc).text = text
                return True
            else:
                raise CardinalityError("ExtractedLicense::text")
        else:
            raise OrderError("ExtractedLicense::text")

    def set_lic_comment(self, doc, comment):
        """
        Set license comment.
        Raise CardinalityError if it is already set.
        Raise OrderError if no license ID defined.
        """
        if self.has_extr_lic(doc):
            if not self.extr_lic_comment_set:
                self.extr_lic_comment_set = True
                self.extr_lic(doc).comment = comment
                return True
            else:
                raise CardinalityError("ExtractedLicense::comment")
        else:
            raise OrderError("ExtractedLicense::comment")


class FileBuilder(rdfbuilders.FileBuilder):
    def __init__(self):
        super(FileBuilder, self).__init__()

    def set_file_checksum(self, doc: Document, checksum: Union[Dict, Checksum, str]) -> bool:
        """
        Set the file checksum.
        checksum - A string
        raise OrderError if no file defined.
        """
        if not self.has_file(doc):
            raise OrderError("No file for checksum defined.")

        if isinstance(checksum, dict):
            algo = checksum.get('algorithm') or 'SHA1'
            identifier = ChecksumAlgorithm.checksum_algorithm_from_string(algo)
            self.file(doc).set_checksum(Checksum(identifier, checksum.get('checksumValue')))
        elif isinstance(checksum, Checksum):
            self.file(doc).set_checksum(checksum)
        elif isinstance(checksum, str):
            self.file(doc).set_checksum(Checksum(ChecksumAlgorithm.SHA1, checksum))
        return True

    def set_file_notice(self, doc, text):
        """
        Set file notice
        Raise OrderError if no file defined.
        Raise CardinalityError if more than one.
        """
        if not self.has_file(doc):
            raise OrderError("File::Notice")

        self.file_notice_set = True
        self.file(doc).notice = text
        return True

    def set_file_type(self, doc, type_value):
        """
        Wrap rdfbuilders.FileBuilder.set_file_type to match the different
        fileType representations.
        This method does not make much sense as it converts the file type (e.g. SOURCE)
        to rdf format (e.g. fileType_source) which is then converted back.
        But this seems to be the kind of workflow that is currently in use here.
        """

        return super(FileBuilder, self).set_file_type(doc, f"namespace#fileType_{type_value.lower()}")

    def set_file_copyright(self, doc, text):
        """
        Raise OrderError if no file defined.
        Raise CardinalityError if more than one.
        """
        if not self.has_file(doc):
            raise OrderError("File::CopyRight")
        if self.file_copytext_set:
            raise CardinalityError("File::CopyRight")
        self.file_copytext_set = True
        self.file(doc).copyright = text
        return True

    def set_file_license_comment(self, doc, text):
        """
        Raise OrderError if no file defined.
        Raise CardinalityError if more than one per file.
        """
        if not self.has_file(doc):
            raise OrderError("File::LicenseComment")
        if self.file_license_comment_set:
            raise CardinalityError("File::LicenseComment")
        self.file(doc).license_comment = text
        return True

    def set_file_attribution_text(self, doc, text):
        """
        Set the file's attribution text.
        """
        if self.has_file(doc):
            self.file(doc).attribution_text = text
            return True

    def set_file_comment(self, doc, text):
        """
        Raise OrderError if no file defined.
        Raise CardinalityError if more than one comment set.
        """
        if not self.has_file(doc):
            raise OrderError("File::Comment")
        if self.file_comment_set:
            raise CardinalityError("File::Comment")
        self.file_comment_set = True
        self.file(doc).comment = text
        return True


class AnnotationBuilder(tagvaluebuilders.AnnotationBuilder):
    def __init__(self):
        super(AnnotationBuilder, self).__init__()

    def add_annotation_comment(self, doc, comment):
        """
        Set the annotation comment.
        Raise CardinalityError if already set.
        Raise OrderError if no annotator defined before.
        """
        if len(doc.annotations) != 0:
            if not self.annotation_comment_set:
                self.annotation_comment_set = True
                doc.annotations[-1].comment = comment
                return True
            else:
                raise CardinalityError("AnnotationComment")
        else:
            raise OrderError("AnnotationComment")


class RelationshipBuilder(tagvaluebuilders.RelationshipBuilder):
    def __init__(self):
        super(RelationshipBuilder, self).__init__()

    def add_relationship_comment(self, doc, comment):
        """
        Set the relationship comment.
        Raise CardinalityError if already set.
        Raise OrderError if no annotator defined before.
        """
        if len(doc.relationships) != 0:
            if not self.relationship_comment_set:
                self.relationship_comment_set = True
                doc.relationships[-1].comment = comment
                return True
            else:
                raise CardinalityError("RelationshipComment")
        else:
            raise OrderError("RelationshipComment")


class Builder(
    DocBuilder,
    CreationInfoBuilder,
    ExternalDocumentRefsBuilder,
    EntityBuilder,
    SnippetBuilder,
    ReviewBuilder,
    LicenseBuilder,
    FileBuilder,
    PackageBuilder,
    AnnotationBuilder,
    RelationshipBuilder,
):
    """
    SPDX document builder.
    """

    def __init__(self):
        super(Builder, self).__init__()
        # FIXME: this state does not make sense
        self.reset()

    def reset(self):
        """
        Reset builder's state for building new documents.
        Must be called between usage with different documents.
        """
        # FIXME: this state does not make sense
        self.reset_creation_info()
        self.reset_document()
        self.reset_package()
        self.reset_file_stat()
        self.reset_reviews()
        self.reset_annotations()
        self.reset_relationship()
        self.reset_extr_lics()



#!/usr/bin/env python

# Parses a tag/value file and writes it out pretty-printed.
# Usage: pp_tv <infile> <outfile>
if __name__ == "__main__":
    import sys
    import codecs
    from spdx.writers.tagvalue import write_document, InvalidDocumentError
    from spdx.parsers.tagvalue import Parser
    from spdx.parsers.loggers import StandardLogger
    from spdx.parsers.loggers import ErrorMessages
    from spdx.parsers.tagvaluebuilders import Builder

    source = sys.argv[1]
    target = sys.argv[2]
    p = Parser(Builder(), StandardLogger())
    p.build()
    with open(source, "r") as f:
        data = f.read()
        document, error = p.parse(data)
        if not error:
            print("Parsing Successful")
            with codecs.open(target, mode="w", encoding="utf-8") as out:
                try:
                    write_document(document, out)
                except InvalidDocumentError:
                    print("Document is Invalid")
                    messages = ErrorMessages()
                    document.validate(messages)
                    print("\n".join(messages.messages))
        else:
            print("Errors encountered while parsing")



# Copyright (c) 2018 Yash M. Nisar
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Tuple, Optional

from spdx import license
from spdx import utils


class Snippet(object):
    """
    Represents an analyzed snippet.
    Fields:
     - spdx_id: Uniquely identify any element in an SPDX document which may be
     referenced  by other elements. Mandatory, one per snippet if the snippet
     is present.
     - name: Name of the snippet. Optional, one. Type: str.
     - comment: General comments about the snippet. Optional, one. Type: str.
     - copyright: Copyright text. Mandatory, one. Type: str.
     - license_comment: Relevant background references or analysis that went
     in to arriving at the Concluded License for a snippet. Optional, one.
     - snip_from_file_spdxid:  Uniquely identify the file in an SPDX document
     which this snippet is associated with. Mandatory, one. Type: str.
     - conc_lics: Contains the license the SPDX file creator has concluded as
     governing the snippet or alternative values if the governing license
     cannot be determined. Mandatory one. Type: license.License or
     utils.NoAssert or utils.SPDXNone.
     - licenses_in_snippet: The list of licenses found in the snippet.
     Mandatory, one or more. Type: license.License or utils.SPDXNone or
     utils.NoAssert.
     - attribution_text: optional string.
     - byte_range: Defines the byte range in the original host file that the
     snippet information applies to. Mandatory, one. Type (int, int)
     - line_range: Defines the line range in the original host file that the
     snippet information applies to. Optional, one. Type (int, int)
    """

    def __init__(
        self, spdx_id=None, copyright=None, snip_from_file_spdxid=None, conc_lics=None, byte_range=None
    ):
        self.spdx_id = spdx_id
        self.name = None
        self.comment = None
        self.copyright = copyright
        self.license_comment = None
        self.attribution_text = None
        self.snip_from_file_spdxid = snip_from_file_spdxid
        self.conc_lics = conc_lics
        self.licenses_in_snippet = []
        self.byte_range: Optional[Tuple[int, int]] = byte_range
        self.line_range: Optional[Tuple[int, int]] = None

    def add_lics(self, lics):
        self.licenses_in_snippet.append(lics)

    def validate(self, messages):
        """
        Validate fields of the snippet and update the messages list with user
        friendly error messages for display.
        """
        self.validate_spdx_id(messages)
        self.validate_copyright_text(messages)
        self.validate_snip_from_file_spdxid(messages)
        self.validate_concluded_license(messages)
        self.validate_licenses_in_snippet(messages)

        return messages

    def validate_spdx_id(self, messages):
        if self.spdx_id is None:
            messages.append("Snippet has no SPDX Identifier.")

    def validate_copyright_text(self, messages):
        if self.copyright and not isinstance(
            self.copyright,
            (str, utils.NoAssert, utils.SPDXNone),
        ):
            messages.append(
                "Snippet copyright must be str or unicode or spdx.utils.NoAssert or spdx.utils.SPDXNone"
            )

    def validate_snip_from_file_spdxid(self, messages):
        if self.snip_from_file_spdxid is None:
            messages.append("Snippet has no Snippet from File SPDX Identifier.")

    def validate_concluded_license(self, messages):
        if self.conc_lics and not isinstance(
            self.conc_lics, (utils.SPDXNone, utils.NoAssert, license.License)
        ):
            messages.append(
                "Snippet concluded license must be instance of "
                "spdx.utils.SPDXNone or spdx.utils.NoAssert or "
                "spdx.license.License"
            )

    def validate_licenses_in_snippet(self, messages):
        for lic in self.licenses_in_snippet:
            if not isinstance(
                lic, (license.License, utils.NoAssert, utils.SPDXNone)
            ):
                messages.append(
                    "License in snippet must be instance of "
                    "spdx.utils.SPDXNone or spdx.utils.NoAssert or "
                    "spdx.license.License"
                )

    def has_optional_field(self, field):
        return bool(getattr(self, field, None))



# Copyright (c) the SPDX tools authors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
import unittest
from collections import OrderedDict

from spdx.parsers import rdf
from spdx.parsers.loggers import StandardLogger
from spdx.parsers.rdfbuilders import Builder as RDFBuilder

from tests import utils_test
from tests.utils_test import TestParserUtils


class TestParser(unittest.TestCase):
    maxDiff = None

    def test_rdf_parser(self, regen=False):
        parser = rdf.Parser(RDFBuilder(), StandardLogger())
        test_file = utils_test.get_test_loc('formats/SPDXRdfExample.rdf', test_data_dir=utils_test.test_data_dir)
        with io.open(test_file, 'rb') as f:
            document, _ = parser.parse(f)
        expected_loc = utils_test.get_test_loc('doc_parse/spdx-expected.json', test_data_dir=utils_test.test_data_dir)
        self.check_document(document, expected_loc, regen=regen)

    def check_document(self, document, expected_loc, regen=False):
        result = TestParserUtils.to_dict(document)

        if regen:
            data = json.dumps(result, indent=2)
            with io.open(expected_loc, 'w') as o:
                o.write(data)

        with io.open(expected_loc, 'r', encoding='utf-8') as ex:
            expected = json.load(ex, object_pairs_hook=OrderedDict)

        assert result == expected



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import json
import os

from spdx.version import Version


_base_dir = os.path.dirname(__file__)
_licenses = os.path.join(_base_dir, "licenses.json")
_exceptions = os.path.join(_base_dir, "exceptions.json")


def _load_list(file_name, object_type="licenses", id_attribute="licenseId"):
    """
    Return a list version tuple and a mapping of licenses
    name->id and id->name loaded from a JSON file
    from https://github.com/spdx/license-list-data
    """
    licenses_map = {}
    with codecs.open(file_name, "rb", encoding="utf-8") as lics:
        licenses = json.load(lics)
    version = tuple(licenses["licenseListVersion"].split("."))
    for lic in licenses[object_type]:
        if lic.get("isDeprecatedLicenseId"):
            continue
        name = lic["name"]
        identifier = lic[id_attribute]
        licenses_map[name] = identifier
        licenses_map[identifier] = name
    return version, licenses_map


def load_license_list(file_name):
    """
    Return the licenses list version tuple and a mapping of licenses
    name->id and id->name loaded from a JSON file
    from https://github.com/spdx/license-list-data
    """
    return _load_list(file_name, object_type="licenses", id_attribute="licenseId")


def load_exception_list(file_name):
    """
    Return the exceptions list version tuple and a mapping of exceptions
    name->id and id->name loaded from a JSON file
    from https://github.com/spdx/license-list-data
    """
    return _load_list(
        file_name, object_type="exceptions", id_attribute="licenseExceptionId"
    )


(_lmajor, _lminor), LICENSE_MAP = load_license_list(_licenses)
LICENSE_LIST_VERSION = Version(major=_lmajor, minor=_lminor)

(_emajor, _eminor), EXCEPTION_MAP = load_exception_list(_exceptions)
EXCEPTION_LIST_VERSION = Version(major=_emajor, minor=_eminor)

assert LICENSE_LIST_VERSION == EXCEPTION_LIST_VERSION
del _emajor, _eminor, EXCEPTION_LIST_VERSION



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import hashlib
import re
from typing import Dict, List, TYPE_CHECKING

from ply import lex
from ply import yacc

from spdx.checksum import ChecksumAlgorithm

if TYPE_CHECKING:
    from spdx.file import File
    from spdx.package import Package
from spdx.relationship import Relationship
from spdx import license


def datetime_iso_format(date):
    """
    Return an ISO-8601 representation of a datetime object.
    """
    return date.isoformat() + "Z"


# Matches an iso 8601 date representation
DATE_ISO_REGEX = re.compile(
    r"(\d\d\d\d)-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)Z", re.UNICODE
)

# Groups for retrieving values from DATE_ISO_REGEX matches.
DATE_ISO_YEAR_GRP = 1
DATE_ISO_MONTH_GRP = 2
DATE_ISO_DAY_GRP = 3
DATE_ISO_HOUR_GRP = 4
DATE_ISO_MIN_GRP = 5
DATE_ISO_SEC_GRP = 6


def datetime_from_iso_format(string):
    """
    Return a datetime object from an iso 8601 representation.
    Return None if string is non conforming.
    """
    match = DATE_ISO_REGEX.match(string)
    if match:
        date = datetime.datetime(
            year=int(match.group(DATE_ISO_YEAR_GRP)),
            month=int(match.group(DATE_ISO_MONTH_GRP)),
            day=int(match.group(DATE_ISO_DAY_GRP)),
            hour=int(match.group(DATE_ISO_HOUR_GRP)),
            second=int(match.group(DATE_ISO_SEC_GRP)),
            minute=int(match.group(DATE_ISO_MIN_GRP)),
        )
        return date
    else:
        return None


class NoAssert(object):
    """
    Represent SPDX NOASSERTION value.
    """

    def to_value(self):
        return "NOASSERTION"

    def __str__(self):
        return self.to_value()

    def __repr__(self):
        return self.to_value()


class UnKnown(object):
    """
    Represent SPDX UNKNOWN value.
    """

    def to_value(self):
        return "UNKNOWN"

    def __str__(self):
        return self.to_value()

    def __repr__(self):
        return self.to_value()

    def __eq__(self, other):
        return self.to_value() == other.to_value()


class SPDXNone(object):
    """
    Represent SPDX None value.
    """

    def to_value(self):
        return "NONE"

    def __str__(self):
        return self.to_value()


class LicenseListLexer(object):
    tokens = ["LP", "RP", "AND", "OR", "LICENSE"]

    def t_LP(self, t):
        r"\("
        return t

    def t_RP(self, t):
        r"\)"
        return t

    def t_AND(self, t):
        r"\s(and|AND)\s"
        t.value = t.value.strip()
        return t

    def t_OR(self, t):
        r"\s(or|OR)\s"
        t.value = t.value.strip()
        return t

    def t_whitespace(self, t):
        r"\s+"
        pass

    def t_LICENSE(self, t):
        r"[A-Za-z.0-9\-+]+"
        t.value = t.value.strip()
        return t

    def t_error(self, t):
        pass

    def input(self, data):
        """Set input, data - str."""
        self.lexer.input(data)

    def token(self):
        """Get the next token or None if exhausted input."""
        return self.lexer.token()

    def build(self, **kwargs):
        """Build lexer, must be called before input or token methods.
        Only need to build once.
        """
        self.lexer = lex.lex(module=self, **kwargs)


class LicenseListParser(object):
    def __init__(self):
        self.lex = LicenseListLexer()
        self.lex.build(reflags=re.UNICODE)
        self.tokens = self.lex.tokens

    def p_disjunction_1(self, p):
        """disjunction : disjunction OR conjunction
        """
        p[0] = license.LicenseDisjunction(p[1], p[3])

    def p_disjunction_2(self, p):
        """disjunction : conjunction
        """
        p[0] = p[1]

    def p_conjunction_1(self, p):
        """conjunction : conjunction AND license_atom
        """
        p[0] = license.LicenseConjunction(p[1], p[3])

    def p_conjunction_2(self, p):
        """conjunction : license_atom
        """
        p[0] = p[1]

    def p_license_atom_1(self, p):
        """license_atom : LICENSE
        """
        p[0] = license.License.from_identifier(p[1])

    def p_license_atom_2(self, p):
        """license_atom : LP disjunction RP
        """
        p[0] = p[2]

    def p_error(self, p):
        pass

    def build(self, **kwargs):
        """Must be called before parse."""
        self.yacc = yacc.yacc(module=self, **kwargs)

    def parse(self, data):
        """Parses a license list and returns a License or None if it failed."""
        try:
            return self.yacc.parse(data, lexer=self.lex)
        except:
            return None


def calc_verif_code(files: List['File']) -> str:
    list_of_file_hashes = []
    hash_algorithm_name = ChecksumAlgorithm.SHA1
    for file in files:
        file_checksum = file.get_checksum(hash_algorithm_name)
        if file_checksum is not None:
            file_checksum_value = file_checksum.value
        else:
            file_checksum_value = file.calculate_checksum(hash_algorithm_name)
        list_of_file_hashes.append(file_checksum_value)

    list_of_file_hashes.sort()

    hasher = hashlib.new(hash_algorithm_name.name.lower())
    hasher.update("".join(list_of_file_hashes).encode("utf-8"))
    return hasher.hexdigest()


def get_files_in_package(package: 'Package', files: List['File'], relationships: List[Relationship]) -> List['File']:
    files_in_package = []
    for file in files:
        if file.spdx_id in [relationship.related_spdx_element for relationship in relationships
                            if relationship.relationship_type == "CONTAINS" and relationship.spdx_element_id == package.spdx_id] \
            or file.spdx_id in [relationship.spdx_element_id for relationship in relationships
                                if relationship.relationship_type == "CONTAINED_BY" and relationship.related_spdx_element == package.spdx_id]:
            files_in_package.append(file)

    return files_in_package


def update_dict_item_with_new_item(current_state: Dict, key: str, item_to_add: str) -> None:
    if key not in current_state:
        current_state[key] = [item_to_add]
    elif item_to_add not in current_state[key]:
        current_state[key].append(item_to_add)



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import zip_longest
from typing import List, TextIO, Tuple, Dict

from spdx import license, utils
from spdx import file as spdx_file
from spdx.document import Document
from spdx.file import File
from spdx.package import Package
from spdx.parsers.loggers import ErrorMessages
from spdx.relationship import Relationship
from spdx.snippet import Snippet
from spdx.version import Version


class InvalidDocumentError(Exception):
    """
    Raised when attempting to write an invalid document.
    """

    pass


def write_separator(out):
    out.write("\n")


def write_separators(out):
    out.write("\n" * 2)


def format_verif_code(package):
    if len(package.verif_exc_files) == 0:
        return package.verif_code
    else:
        return "{0} ({1})".format(package.verif_code, ",".join(package.verif_exc_files))


def write_value(tag, value, out):
    out.write("{0}: {1}\n".format(tag, value))


def write_range(tag, value, out):
    out.write("{0}: {1}:{2}\n".format(tag, value[0], value[1]))


def write_text_value(tag, value, out):
    if "\n" in value:
        out.write("{0}: <text>{1}</text>\n".format(tag, value))
    else:
        write_value(tag, value, out)


def write_creation_info(creation_info, out):
    """
    Write the creation info to out.
    """
    out.write("# Creation Info\n\n")
    # Write sorted creators
    for creator in sorted(creation_info.creators):
        write_value("Creator", creator, out)

    # write created
    write_value("Created", creation_info.created_iso_format, out)
    # possible comment
    if creation_info.has_comment:
        write_text_value("CreatorComment", creation_info.comment, out)


def write_review(review, out):
    """
    Write the fields of a single review to out.
    """
    write_value("Reviewer", review.reviewer, out)
    write_value("ReviewDate", review.review_date_iso_format, out)
    if review.has_comment:
        write_text_value("ReviewComment", review.comment, out)


def write_annotation(annotation, out):
    """
    Write the fields of a single annotation to out.
    """
    write_value("Annotator", annotation.annotator, out)
    write_value("AnnotationDate", annotation.annotation_date_iso_format, out)
    if annotation.has_comment:
        write_text_value("AnnotationComment", annotation.comment, out)
    write_value("AnnotationType", annotation.annotation_type, out)
    if annotation.spdx_id:
        write_value("SPDXREF", annotation.spdx_id, out)


def write_relationship(relationship_term, out):
    """
    Write the fields of relationships to out.
    """
    write_value("Relationship", relationship_term.relationship, out)
    if relationship_term.has_comment:
        write_text_value(
            "RelationshipComment", relationship_term.relationship_comment, out
        )


def write_file_type(ftype, out):
    write_value("FileType", ftype, out)


def write_file(spdx_file, out):
    """
    Write a file fields to out.
    """
    out.write("# File\n\n")
    write_value("FileName", spdx_file.name, out)
    if spdx_file.spdx_id:
        write_value("SPDXID", spdx_file.spdx_id, out)
    for file_type in spdx_file.file_types:
        write_file_type(file_type.name, out)
    for file_checksum in spdx_file.checksums.values():
        write_value("FileChecksum", file_checksum.to_tv(), out)
    if spdx_file.has_optional_field("conc_lics"):
        if isinstance(
            spdx_file.conc_lics, (license.LicenseConjunction, license.LicenseDisjunction)
        ):
            write_value("LicenseConcluded", "({0})".format(spdx_file.conc_lics), out)
        else:
            write_value("LicenseConcluded", spdx_file.conc_lics, out)

    # remove duplicates
    lics_in_file = []
    [lics_in_file.append(x) for x in spdx_file.licenses_in_file if x not in lics_in_file]
    # write sorted list
    for lics in sorted(lics_in_file):
        write_value("LicenseInfoInFile", lics, out)

    if spdx_file.has_optional_field("copyright"):
        if isinstance(spdx_file.copyright, str):
            write_text_value("FileCopyrightText", spdx_file.copyright, out)
        else:
            write_value("FileCopyrightText", spdx_file.copyright, out)

    if spdx_file.has_optional_field("license_comment"):
        write_text_value("LicenseComments", spdx_file.license_comment, out)

    if spdx_file.has_optional_field("attribution_text"):
        write_text_value("FileAttributionText", spdx_file.attribution_text, out)

    if spdx_file.has_optional_field("comment"):
        write_text_value("FileComment", spdx_file.comment, out)

    if spdx_file.has_optional_field("notice"):
        write_text_value("FileNotice", spdx_file.notice, out)

    for contributor in sorted(spdx_file.contributors):
        write_value("FileContributor", contributor, out)

    for dependency in sorted(spdx_file.dependencies):
        write_value("FileDependency", dependency, out)

    names = spdx_file.artifact_of_project_name
    homepages = spdx_file.artifact_of_project_home
    uris = spdx_file.artifact_of_project_uri

    for name, homepage, uri in sorted(zip_longest(names, homepages, uris)):
        write_value("ArtifactOfProjectName", name, out)
        if homepage is not None:
            write_value("ArtifactOfProjectHomePage", homepage, out)
        if uri is not None:
            write_value("ArtifactOfProjectURI", uri, out)


def write_snippet(snippet, out):
    """
    Write snippet fields to out.
    """
    out.write("# Snippet\n\n")
    write_value("SnippetSPDXID", snippet.spdx_id, out)
    write_value("SnippetFromFileSPDXID", snippet.snip_from_file_spdxid, out)
    if snippet.has_optional_field("copyright"):
        write_text_value("SnippetCopyrightText", snippet.copyright, out)
    write_range("SnippetByteRange", snippet.byte_range, out)
    if snippet.has_optional_field("line_range"):
        write_range("SnippetLineRange", snippet.line_range, out)
    if snippet.has_optional_field("name"):
        write_value("SnippetName", snippet.name, out)
    if snippet.has_optional_field("comment"):
        write_text_value("SnippetComment", snippet.comment, out)
    if snippet.has_optional_field("license_comment"):
        write_text_value("SnippetLicenseComments", snippet.license_comment, out)
    if snippet.has_optional_field("attribution_text"):
        write_text_value("SnippetAttributionText", snippet.attribution_text, out)
    if snippet.has_optional_field("conc_lics"):
        if isinstance(
            snippet.conc_lics, (license.LicenseConjunction, license.LicenseDisjunction)
        ):
            write_value("SnippetLicenseConcluded", "({0})".format(snippet.conc_lics), out)
        else:
            write_value("SnippetLicenseConcluded", snippet.conc_lics, out)
    # Write sorted list
    for lics in sorted(snippet.licenses_in_snippet):
        write_value("LicenseInfoInSnippet", lics, out)


def write_package(package, out):
    """
    Write a package fields to out.
    """
    out.write("# Package\n\n")
    if package.name:
        write_value("PackageName", package.name, out)
    if package.spdx_id:
        write_value("SPDXID", package.spdx_id, out)
    if package.has_optional_field("version"):
        write_value("PackageVersion", package.version, out)
    write_value("PackageDownloadLocation", package.download_location, out)

    if package.has_optional_field("files_analyzed"):
        write_value("FilesAnalyzed", package.files_analyzed, out)

    if package.has_optional_field("summary"):
        write_text_value("PackageSummary", package.summary, out)

    if package.has_optional_field("attribution_text"):
        write_text_value("PackageAttributionText", package.attribution_text, out)

    if package.has_optional_field("source_info"):
        write_text_value("PackageSourceInfo", package.source_info, out)

    if package.has_optional_field("file_name"):
        write_value("PackageFileName", package.file_name, out)

    if package.has_optional_field("supplier"):
        write_value("PackageSupplier", package.supplier, out)

    if package.has_optional_field("originator"):
        write_value("PackageOriginator", package.originator, out)

    for package_checksum in package.checksums.values():
        write_value("PackageChecksum", package_checksum.to_tv(), out)

    if package.has_optional_field("verif_code"):
        write_value("PackageVerificationCode", format_verif_code(package), out)

    if package.has_optional_field("description"):
        write_text_value("PackageDescription", package.description, out)

    if package.has_optional_field("comment"):
        write_text_value("PackageComment", package.comment, out)

    if package.has_optional_field("license_declared"):
        if isinstance(
            package.license_declared,
            (license.LicenseConjunction, license.LicenseDisjunction),
        ):
            write_value(
                "PackageLicenseDeclared", "({0})".format(package.license_declared), out
            )
        else:
            write_value("PackageLicenseDeclared", package.license_declared, out)

    if package.has_optional_field("conc_lics"):
        if isinstance(
            package.conc_lics, (license.LicenseConjunction, license.LicenseDisjunction)
        ):
            write_value("PackageLicenseConcluded", "({0})".format(package.conc_lics), out)
        else:
            write_value("PackageLicenseConcluded", package.conc_lics, out)

    # Write sorted list of licenses.
    for lics in sorted(package.licenses_from_files):
        write_value("PackageLicenseInfoFromFiles", lics, out)

    if package.has_optional_field("license_comment"):
        write_text_value("PackageLicenseComments", package.license_comment, out)

    if package.has_optional_field("cr_text"):
        if isinstance(package.cr_text, str):
            write_text_value("PackageCopyrightText", package.cr_text, out)
        else:
            write_value("PackageCopyrightText", package.cr_text, out)

    if package.has_optional_field("homepage"):
        write_value("PackageHomePage", package.homepage, out)

    for pkg_ref in package.pkg_ext_refs:
        pkg_ref_str = " ".join(
            [pkg_ref.category, pkg_ref.pkg_ext_ref_type, pkg_ref.locator]
        )
        write_value("ExternalRef", pkg_ref_str, out)
        if pkg_ref.comment:
            write_text_value("ExternalRefComment", pkg_ref.comment, out)

    if package.has_optional_field("primary_package_purpose"):
        write_value("PrimaryPackagePurpose", package.primary_package_purpose.name.replace("_", "-"), out)

    if package.has_optional_field("built_date"):
        write_value("BuiltDate", utils.datetime_iso_format(package.built_date), out)

    if package.has_optional_field("release_date"):
        write_value("ReleaseDate", utils.datetime_iso_format(package.release_date), out)

    if package.has_optional_field("valid_until_date"):
        write_value("ValidUntilDate", utils.datetime_iso_format(package.valid_until_date), out)


def write_extracted_licenses(lics, out):
    """
    Write extracted licenses fields to out.
    """
    write_value("LicenseID", lics.identifier, out)

    if lics.full_name is not None:
        write_value("LicenseName", lics.full_name, out)

    if lics.comment is not None:
        write_text_value("LicenseComment", lics.comment, out)

    for xref in sorted(lics.cross_ref):
        write_value("LicenseCrossReference", xref, out)

    write_text_value("ExtractedText", lics.text, out)


def write_document(document, out, validate=True):
    """
    Write an SPDX tag value document.
    - document - spdx.document instance.
    - out - file like object that will be written to.
    Optionally `validate` the document before writing and raise
    InvalidDocumentError if document.validate returns False.
    """
    messages = ErrorMessages()
    messages = document.validate(messages)
    if validate and messages:
        raise InvalidDocumentError(messages)

    # Write out document information
    out.write("# Document Information\n\n")
    write_value("SPDXVersion", str(document.version), out)
    write_value("DataLicense", document.data_license.identifier, out)
    if document.namespace:
        write_value("DocumentNamespace", document.namespace, out)
    if document.name:
        write_value("DocumentName", document.name, out)
    if document.creation_info.license_list_version:
        version: Version = document.creation_info.license_list_version
        write_value("LicenseListVersion", str(version.major) + "." + str(version.minor), out)
    write_value("SPDXID", "SPDXRef-DOCUMENT", out)
    if document.has_comment:
        write_text_value("DocumentComment", document.comment, out)
    for doc_ref in document.ext_document_references:
        doc_ref_str = " ".join(
            [
                doc_ref.external_document_id,
                doc_ref.spdx_document_uri,
                doc_ref.checksum.identifier.name + ": " + doc_ref.checksum.value,
            ]
        )
        write_value("ExternalDocumentRef", doc_ref_str, out)
    write_separators(out)
    # Write out creation info
    write_creation_info(document.creation_info, out)
    write_separators(out)

    # Write sorted reviews
    if document.reviews:
        out.write("# Reviews\n\n")
        for review in sorted(document.reviews):
            write_review(review, out)
            write_separator(out)
        write_separator(out)

    # Write sorted annotations
    if document.annotations:
        out.write("# Annotations\n\n")
        for annotation in sorted(document.annotations):
            write_annotation(annotation, out)
            write_separator(out)
        write_separator(out)

    relationships_to_write, contained_files_by_package_id = scan_relationships(document.relationships,
                                                                               document.packages, document.files)
    contained_snippets_by_file_id = determine_files_containing_snippets(document.snippet, document.files)
    packaged_file_ids = [file.spdx_id for files_list in contained_files_by_package_id.values()
                         for file in files_list]
    filed_snippet_ids = [snippet.spdx_id for snippets_list in contained_snippets_by_file_id.values()
                         for snippet in snippets_list]

    # Write Relationships
    if relationships_to_write:
        out.write("# Relationships\n\n")
        for relationship in relationships_to_write:
            write_relationship(relationship, out)
        write_separators(out)

    # Write snippet info
    for snippet in document.snippet:
        if snippet.spdx_id not in filed_snippet_ids:
            write_snippet(snippet, out)
            write_separators(out)

    # Write file info
    for file in document.files:
        if file.spdx_id not in packaged_file_ids:
            write_file(file, out)
            write_separators(out)
            if file.spdx_id in contained_snippets_by_file_id:
                write_snippets(contained_snippets_by_file_id[file.spdx_id], out)

    # Write out package info
    for package in document.packages:
        write_package(package, out)
        write_separators(out)
        if package.spdx_id in contained_files_by_package_id:
            for file in contained_files_by_package_id[package.spdx_id]:
                write_file(file, out)
                write_separators(out)
                if file.spdx_id in contained_snippets_by_file_id:
                    write_snippets(contained_snippets_by_file_id[file.spdx_id], out)
                    break

    if document.extracted_licenses:
        out.write("# Extracted Licenses\n\n")
        for lic in sorted(document.extracted_licenses):
            write_extracted_licenses(lic, out)
            write_separator(out)
        write_separator(out)


def write_snippets(snippets_to_write: List, out: TextIO) -> None:
    for snippet in snippets_to_write:
        write_snippet(snippet, out)
        write_separators(out)


def scan_relationships(relationships: List[Relationship], packages: List[Package], files: List[File]) \
        -> Tuple[List, Dict]:
    contained_files_by_package_id = dict()
    relationships_to_write = []
    files_by_spdx_id = {file.spdx_id: file for file in files}
    packages_spdx_ids = [package.spdx_id for package in packages]
    for relationship in relationships:
        if relationship.relationship_type == "CONTAINS" and \
            relationship.spdx_element_id in packages_spdx_ids and \
                relationship.related_spdx_element in files_by_spdx_id.keys():
            contained_files_by_package_id.setdefault(relationship.spdx_element_id, []).append(
                files_by_spdx_id[relationship.related_spdx_element])
            if relationship.has_comment:
                relationships_to_write.append(relationship)
        elif relationship.relationship_type == "CONTAINED_BY" and \
            relationship.related_spdx_element in packages_spdx_ids and \
                relationship.spdx_element_id in files_by_spdx_id:
            contained_files_by_package_id.setdefault(relationship.related_spdx_element, []).append(
                files_by_spdx_id[relationship.spdx_element_id])
            if relationship.has_comment:
                relationships_to_write.append(relationship)
        else:
            relationships_to_write.append(relationship)

    return relationships_to_write, contained_files_by_package_id


def determine_files_containing_snippets(snippets: List[Snippet], files: List[File]) -> Dict:
    contained_snippets_by_file_id = dict()
    for snippet in snippets:
        if snippet.snip_from_file_spdxid in [file.spdx_id for file in files]:
            contained_snippets_by_file_id.setdefault(snippet.snip_from_file_spdxid, []).append(snippet)

    return contained_snippets_by_file_id



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from functools import total_ordering

from spdx import config
from spdx import utils
from spdx.version import Version


@total_ordering
class Creator(object):
    """
    Creator entity.
    Fields:
    - name: creator's name/identifier
    """

    def __init__(self, name):
        self.name = name

    # FIXME: do not override eq and not hash
    def __eq__(self, other):
        return isinstance(other, Creator) and self.name == other.name

    def __lt__(self, other):
        return isinstance(other, Creator) and self.name < other.name


@total_ordering
class Organization(Creator):
    """
    Organization entity.
    Fields:
    - name: Org's name/identifier. Mandatory. Type: str.
    - email: Org's email address. Optional. Type: str.
    """

    def __init__(self, name, email=None):
        super(Organization, self).__init__(name)
        self.email = email

    # FIXME: do not override eq and not hash
    def __eq__(self, other):
        return isinstance(other, Organization) and (self.name, self.email) == (
            other.name,
            other.email,
        )

    def __lt__(self, other):
        return isinstance(other, Organization) and (self.name, self.email) < (
            other.name,
            other.email,
        )

    def to_value(self):
        if self.email:
            return "Organization: {0} ({1})".format(self.name, self.email)
        else:
            return "Organization: {0}".format(self.name)

    def __str__(self):
        return self.to_value()


@total_ordering
class Person(Creator):
    """
    Person entity.
    Fields:
    - name: person's name/identifier. Mandatory. Type: str.
    - email: person's email address. Optional. Type: str.
    """

    def __init__(self, name, email=None):
        super(Person, self).__init__(name)
        self.email = email

    # FIXME: do not override eq and not hash
    def __eq__(self, other):
        return isinstance(other, Person) and (self.name, self.email) == (
            other.name,
            other.email,
        )

    def __lt__(self, other):
        return isinstance(other, Person) and (self.name, self.email) < (
            other.name,
            other.email,
        )

    def to_value(self):
        if self.email is not None:
            return "Person: {0} ({1})".format(self.name, self.email)
        else:
            return "Person: {0}".format(self.name)

    def __str__(self):
        return self.to_value()


class Tool(Creator):
    """
    Tool entity.
    Fields:
    - name: tool identifier, with version. Type: str.
    """

    def __init__(self, name):
        super(Tool, self).__init__(name)

    def to_value(self):
        return "Tool: {0}".format(self.name)

    def __str__(self):
        return self.to_value()


class CreationInfo(object):
    """
    Represent a document creation info.
    Fields:
    - creators: List of creators. At least one required.
        Type: Creator.
    - comment: Creation comment, optional. Type: str.
    - license_list_version: version of SPDX license used in creation of SPDX
        document. One, optional. Type: spdx.version.Version
    - created: Creation date. Mandatory one. Type: datetime.
    """

    def __init__(
        self,
        created=None,
        comment=None,
        license_list_version=config.LICENSE_LIST_VERSION,
    ):
        self.creators = []
        self.created = created
        self.comment = comment
        self.license_list_version = license_list_version

    def add_creator(self, creator):
        self.creators.append(creator)

    def remove_creator(self, creator):
        self.creators.remove(creator)

    def set_created_now(self):
        self.created = datetime.utcnow().replace(microsecond=0)

    def set_license_list_version(self, license_list_version):
        self.license_list_version = Version.from_str(license_list_version)

    @property
    def created_iso_format(self):
        return utils.datetime_iso_format(self.created)

    @property
    def has_comment(self):
        return self.comment is not None

    def validate(self, messages):
        """
        Check that all the fields are valid.
        Appends any error messages to messages parameter shall be a ErrorMessages.
        """
        self.validate_creators(messages)
        self.validate_created(messages)

    def validate_creators(self, messages):
        if len(self.creators) == 0:
            messages.append("No creators defined, must have at least one.")

    def validate_created(self, messages):
        if self.created is None:
            messages.append("Creation info missing created date.")
